# coding: utf-8
"""

    ファイル名：   SimulatedExtrapolation.py

    処理内容：

        ゼロ濃度外挿シミュレーションのプロット

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF

"""
import sys
import os
import numpy    as np
import matplotlib.pyplot        as plt
from mpl_toolkits.mplot3d       import Axes3D
from matplotlib.backends.backend_tkagg  import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter                 import Tk, Dialog
from molass_legacy.KekLib.BasicUtils                 import clear_dirs_with_retry, get_caller_module
from molass_legacy.KekLib.TkSupplements              import tk_set_icon_portable
from molass_legacy.KekLib.NumpyUtils                 import np_savetxt
from SphericalModel             import XrayScattering
import OurStatsModels           as sm
from SerialAtsasTools           import AlmergeExecutor

this_dir = os.path.dirname( os.path.abspath( __file__ ) )

SLICE_NUM_POINTS = 400

LOG_PLOT = False
LOG_PLOT_A = True

class SimulatedExtrapolation:
    def __init__( self ):
        self.work_dir = os.path.abspath( this_dir + '/../../work' )
        self.data_dir = self.work_dir + '/data'
        clear_dirs_with_retry( [ self.work_dir, self.data_dir ] )

        self.generate_data()

    def __del__( self ):
        clear_dirs_with_retry( [ self.work_dir ] )

    def generate_data( self ):
        self.x = x = np.linspace( 0.005, 0.8, 800 )
        C = 1e-2
        self.cvector_all = 0.1 * np.exp( - C * ( np.arange( 50 ) - 25 )**2  )
        self.indeces = np.arange( 0, 25 )

        intensity_list = []
        d_error_list = []
        filepaths = []
        for i in self.indeces:
            c = self.cvector_all[i]
            intensity = XrayScattering( c, 29 )
            if len(sys.argv) == 1:
                noise_level = 1/( 0.01 + c)
                ii = intensity( x ) * ( 1 + ( noise_level * 1e-4 + ( x - 0.05 ) **2 )* np.random.normal( 0, noise_level, len(x) ) )
                e = np.ones( len(x) ) * 1e-5 * noise_level
            else:
                noise_level = 1/c
                ii = intensity( x ) * ( 1 + ( x**2 )* np.random.normal( 0, noise_level, len(x) ) )
                e = np.ones( len(x) ) * 1e-6 * noise_level
            intensity_list.append( ii )
            d_error_list.append( e )
            filepath = self.data_dir + '/' +'SAMPLE-%04d.dat' % i
            data = np.array( [ x, ii, e ] ).T
            np_savetxt( filepath, data )
            filepaths.append( filepath )

        self.intensity_array = np.array( intensity_list )
        self.d_error_array = np.array( d_error_list )
        self.filepaths = filepaths
        # print( 'intensity_array.shape=', self.intensity_array.shape )
        # print( 'd_error_array.shape=', self.d_error_array.shape )

        self.x_slice = slice( 0, SLICE_NUM_POINTS )
        self.x_ = self.x[self.x_slice]

        self.x_i = 50
        self.x_f = self.x_[self.x_i]
        self.cvector_ = self.cvector_all[self.indeces]

        self.x1_ = self.cvector_
        self.x2_ = self.cvector_**2
        self.X = np.array( [ self.x1_, self.x2_] ).T

        param_list = []
        error_list = []
        for i in range(0,SLICE_NUM_POINTS):
            z_f = self.intensity_array[self.indeces,i]
            e2  = self.d_error_array[self.indeces,i]**2
            w   = 1/e2
            model   = sm.WLS(z_f, self.X, weights=w)
            result  = model.fit()
            param_list.append( result.params )
            # error_list.append( np.diag( result.cov_params() ) )
            error = np.sqrt( np.dot( model.XtWX_inv_XtW**2, e2 ) )
            error_list.append( error )

        self.params_array = np.array( param_list )
        self.error_array = np.array( error_list )

    def plot_concentration_elution( self, ax ):
        ax.set_title( 'Concentration Curve' )
        ax.set_xlabel( 'Elution No.' )
        ax.set_ylabel( 'Concentration' )
        ax.plot( self.indeces, self.cvector_ )

    def plot_xray_scattering( self, ax ):
        ax.set_title( 'Scattering Curves on Q[0:200] and the plane at Q[%d]=%.3g' % (self.x_i, self.x_f) )
        ax.set_xlabel( 'Q' )
        ax.set_ylabel( 'Elution No.' )
        zlabel_ = 'ln(Intensity)' if LOG_PLOT else 'Intensity'
        ax.set_zlabel( zlabel_ )

        x_ = self.x_

        for i in self.indeces:
            c = self.cvector_all[i]
            # y_ = np.ones( len(x_) ) * c
            y_ = np.ones( len(x_) ) * i
            if LOG_PLOT:
                z_ = np.log( self.intensity_array[i,self.x_slice] )
            else:
                z_ = self.intensity_array[i,self.x_slice]
            ax.plot( x_, y_, z_ )

        ymin, ymax = ax.get_ylim()
        ax.set_ylim( ymin, ymax )
        zmin, zmax = ax.get_zlim()
        ax.set_zlim( zmin, zmax )
        ax.plot( np.ones(5)*self.x_f,
                    [ymin, ymax, ymax, ymin, ymin ],
                    [zmin, zmin, zmax, zmax, zmin ],
                    ':', color='black' )
        if LOG_PLOT:
            z_i = np.log( self.intensity_array[self.indeces,self.x_i] )
        else:
            z_i = self.intensity_array[self.indeces,self.x_i]
        ax.plot( np.ones(len(self.cvector_))*self.x_f, self.indeces, z_i, ':', color='red' )

    def plot_xray_scattering_elution( self, ax ):
        ax.set_title( 'Scattering Variation with errorbar at Q[%d]=%.3g' % (self.x_i, self.x_f) )
        ax.set_xlabel( 'Seq No.' )
        ax.set_ylabel( 'Concentration' )
        y_f = self.intensity_array[self.indeces,self.x_i]
        y_e = self.d_error_array[self.indeces,self.x_i]
        ax.plot( self.indeces, y_f, ':', color='red' )
        ax.errorbar( self.indeces, y_f, yerr=y_e, color='blue', fmt='none' )

    def plot_regression( self, ax ):
        ax.set_title( 'Multivariate Regression at Q[%d]=%.3g' % (self.x_i, self.x_f) )
        ax.set_xlabel( 'C' )
        ax.set_ylabel( 'C²' )
        ax.set_zlabel( 'Scattering' )

        params = self.params_array[self.x_i,:]
        z_f = self.intensity_array[self.indeces,self.x_i]
        z_r = np.dot( self.X, params )

        ax.scatter( self.x1_, self.x2_, z_f )
        ax.scatter( self.x1_, self.x2_, z_r, color='orange' )

        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
        ax.set_xlim( xmin, xmax )
        ax.set_ylim( ymin, ymax )

        xx, yy = np.meshgrid( np.linspace(xmin, xmax, 10), np.linspace(ymin, ymax, 10) )
        zz = xx * params[0] + yy * params[1]
        ax.plot_surface(xx, yy, zz, alpha=0.1 )

    def plot_regression_params( self, ax_a, ax_b ):
        ax_a.set_title( 'Extrapolated A(q)' )
        ax_a.set_xlabel( 'Q' )
        ylabel_ = 'ln( Intensity / C )' if LOG_PLOT_A else 'Intensity / C'
        ax_a.set_ylabel( ylabel_ )

        ax_b.set_title( 'Extrapolated B(q)' )
        ax_b.set_xlabel( 'Q' )
        ax_b.set_ylabel( 'Intensity / C²' )

        A = self.params_array[:,0]
        B = self.params_array[:,1]
        A_ = np.log(A) if LOG_PLOT_A else A
        x_ = self.x_
        x_f = self.x_f
        x_i = self.x_i

        ax_a.plot( x_, A_, color='red', label='SA A(q)' )
        ax_b.plot( x_, B, color='red', label='SA B(q)' )
        ax_a.errorbar( x_f, A[x_i], yerr=self.error_array[x_i,0], color='blue', fmt='none' )
        ax_b.errorbar( x_f, B[x_i], yerr=self.error_array[x_i,1], color='blue', fmt='none' )

        c = 1
        intensity = XrayScattering( c, 29 )
        At = intensity.term1( x_ )
        Bt = -At*intensity.term2( x_ )

        ax_a.plot( x_, np.log(At), color='cyan', label='True A(q)' )
        ax_b.plot( x_, Bt, color='cyan', label='True B(q)' )

        almerge = AlmergeExecutor()
        exz_file = self.work_dir + '/exz_data.dat'
        almerge_result = almerge.execute( self.cvector_, self.filepaths, self.indeces, exz_file )
        max_c  = np.max( self.cvector_ )
        Aa = almerge_result.exz_array[self.x_slice,1] / max_c

        ax_a.plot( x_, np.log(Aa), color='gray', label='Almerge A(q)' )

        # np_savetxt( 'A.csv', A )
        # np_savetxt( 'B.csv', B )

        for ax in [ax_a, ax_b]:
            ymin, ymax = ax.get_ylim()
            ax.set_ylim( ymin, ymax )
            ax.plot( [ x_f, x_f ], [ymin, ymax], ':', color='black' )

        for ax in [ax_a, ax_b]:
            ax.legend()

class SimulatedExtrapolationDialog( Dialog ):
    def __init__( self, title, parent=None ):
        self.grab = 'local'     # used in grab_set
        self.parent     = parent
        self.title_     = title
        self.applied    = None
        self.caller_module = get_caller_module( level=2 )

    def show( self, figsize=None, message=None ):
        self.figsize    = figsize
        self.message    = message
        self.parent.config( cursor='wait' )
        self.parent.update()

        Dialog.__init__( self, self.parent, self.title_ )
        # TODO: self.resizable(width=False, height=False)

    def body( self, body_frame ):   # overrides parent class method
        tk_set_icon_portable( self, module=self.caller_module )

        if self.message is not None:
            self.msg = Tk.Label( body_frame, text=self.message, bg='white' )
            self.msg.pack( fill=Tk.BOTH, expand=1, pady=20 )
            # msg.insert( Tk.INSERT, self.message )
            # msg.config( state=Tk.DISABLED )

        cframe = Tk.Frame( body_frame )
        cframe.pack()

        figsize_ = ( 20, 11 ) if self.figsize is None else self.figsize

        fig = plt.figure( figsize=figsize_ )

        self.mpl_canvas = FigureCanvasTkAgg( fig, cframe )
        self.mpl_canvas.draw()
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )

        # it seems that draw_func should be called after the creation of mpl_canvas
        # in order to enable 3d-rotation of the figure
        self.draw( fig )
        self.parent.config( cursor='' )

        self.protocol( "WM_DELETE_WINDOW", self.ok )

    def buttonbox( self, frame=None ):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)

    def draw( self, fig ):
        se = SimulatedExtrapolation()

        ax1 = fig.add_subplot( 231 )
        ax2 = fig.add_subplot( 232, projection='3d' )
        ax3 = fig.add_subplot( 233 )
        ax4 = fig.add_subplot( 234, projection='3d' )
        ax5 = fig.add_subplot( 235 )
        ax6 = fig.add_subplot( 236 )

        se.plot_concentration_elution( ax1 )
        se.plot_xray_scattering( ax2 )
        se.plot_xray_scattering_elution( ax3 )
        se.plot_regression( ax4 )
        se.plot_regression_params( ax5, ax6 )

        fig.tight_layout()



import arviz as az
import matplotlib.pyplot as plt
import numpy as np
import pymc as pm
import pytensor.tensor as pt
from pytensor.tensor import TensorVariable


def my_model1(theta, freq):
    d_c1, a_c = theta

    d1 = 0.0075
    d2 = 0.0625

    LossE_list = []
    for k in range(0, len(freq)):
        f = freq[k]
        LossE = (f**2)*d1*d_c1 + (f)*d2*(a_c**2)
        LossE_list.append(LossE)

    stacked_LossE = np.array(LossE_list)

    return stacked_LossE

def my_loglike1(theta, freq, data, sigma):
    model = my_model1(theta, freq)
    return -(0.5 / sigma**2) * np.sum((data - model) ** 2)

# define a pytensor Op for our likelihood function
class LogLike(pt.Op):

    itypes = [pt.dvector]  # expects a vector of parameter values when called
    otypes = [pt.dscalar]  # outputs a single scalar value (the log likelihood)

    def __init__(self, loglike, data, freq, sigma):

        # add inputs as class attributes
        self.likelihood = loglike
        self.data = data
        self.freq = freq
        self.sigma = sigma

    def perform(self, node, inputs, outputs):
        # the method that is used when calling the Op
        (theta,) = inputs  # this will contain my variables

        # call the log-likelihood function
        logl = self.likelihood(theta, self.freq, self.data, self.sigma)

        outputs[0][0] = np.array(logl)  # output the log-likelihood

freq = np.arange(1,2000,5)
N = len(freq)
sigma = 0.05  # standard deviation of noise
d_c1_true = 0.003  # true gradient
a_c_true = 0.04  # true y-intercept

truemodel = my_model1([d_c1_true, a_c_true], freq)

# make data
rng = np.random.default_rng(716743)
data = sigma * rng.normal(size=N) + truemodel

# create our Op
logl1 = LogLike(my_loglike1, data, freq, sigma)

def logp(value: TensorVariable, mu: TensorVariable) -> TensorVariable:
    return -(value - mu)**2

if __name__ == '__main__':
    
    with pm.Model():
        # uniform priors 
        d_c1 = pm.Uniform('d_c1_model1',lower=0.001, upper=0.009)
        a_c = pm.Uniform("a_c_model1", lower=0.015, upper=0.09)

        mu = pm.Normal('mu',0,1)

        # convert to a tensor vector
        theta = pt.as_tensor_variable([d_c1, a_c])

        # use a Potential to "call" the Op and include it in the logp computation
        pm.Potential("likelihood", logl1(theta))

        pm.CustomDist('custom_dist',mu, logp=logp, observed=data)

        # Use custom number of draws to replace the HMC based defaults
        idata_mh = pm.smc.sample_smc(draws=100, cores=10, chains=4, return_inferencedata=True, idata_kwargs=dict(log_likelihood=True))


"""
    MatplotlibUtils.py

    Collected 2023-2024, SAXS Team, KEK-PF  
"""
from mpl_toolkits.mplot3d.proj3d import proj_transform
from mpl_toolkits.mplot3d.axes3d import Axes3D
from matplotlib.text import Annotation

"""
    c.f.
        https://gist.github.com/WetHat/1d6cd0f7309535311a539b42cccca89c
        https://stackoverflow.com/questions/10374930/matplotlib-annotating-a-3d-scatter-plot
"""
class Annotation3D(Annotation):

    def __init__(self, text, xyz, *args, **kwargs):
        super().__init__(text, xy=(0, 0), *args, **kwargs)
        self._xyz = xyz

    def draw(self, renderer):
        x2, y2, z2 = proj_transform(*self._xyz, self.axes.M)
        self.xy = (x2, y2)
        super().draw(renderer)

def _annotate3D(ax, text, xyz, *args, **kwargs):
    '''Add anotation `text` to an `Axes3d` instance.'''

    annotation = Annotation3D(text, xyz, *args, **kwargs)
    ax.add_artist(annotation)

setattr(Axes3D, 'annotate3D', _annotate3D)

"""
    c.f.
        https://stackoverflow.com/questions/44465242/getting-the-legend-label-for-a-line-in-matplotlib
"""
def get_labeltext_from_line(line):
    leg = line.axes.get_legend()
    for h, t in zip(leg.legendHandles, leg.texts):
        if h.get_label() == line.get_label():
            return t    # changed from t.get_text() the above reference code


# -*- coding: utf-8 -*-
"""

    ファイル名：   GuiPreferences.py

    処理内容：

        GUI のオプション設定情報

"""
from __future__ import division, print_function, unicode_literals

from molass_legacy.KekLib.OurTkinter     import Tk, Dialog, ttk, tk_set_icon_portable
import OurMessageBox    as MessageBox
from Preferences    import get_preference, set_preference, save_preferences

class PreferencesDialog( Dialog ):

    def __init__( self, parent, title, pilatus_counter, action=0 ):
        self.grab = 'local'     # used in grab_set
        self.pilatus_counter    = pilatus_counter
        self.action             = action
        self.applied            = False
        self.couter_id_changed  = False
        self.adj_output_changed = False
        self.trace_             = True

        Dialog.__init__(self, parent, title) # this calls body

    def body( self, body_frame ):   # overrides parent class method

        tk_set_icon_portable( self, 'synthesizer' )

        iframe = Tk.Frame( body_frame );
        iframe.pack( expand=1, fill=Tk.BOTH, padx=10, pady=10 )

        label_spacing       = '     '
        text_entry_width    = 12

        grid_row = -1

        if self.action in [ 0, 3 ]:
            grid_row += 1
            syn_method_label = Tk.Label( iframe, text= 'Sythesizing Method:' + label_spacing )
            syn_method_label.grid( row=grid_row, column=0, sticky=Tk.E )
            self.syn_method = Tk.StringVar()
            self.syn_method.set( get_preference( 'syn_method' ) )
            j = 0
            for t, method in ( [ 'Just Cover', 'cover' ], [ 'Compute Average', 'average']  ):
                b = Tk.Radiobutton( iframe, text='%-22.22s' % (t),
                                variable=self.syn_method, value=method )
                j += 1
                b.grid( row=grid_row, column=j, sticky=Tk.W )

            grid_row += 1
            self.detection_counter = Tk.StringVar()
            self.detection_counter.set( get_preference( 'detection_counter' ) )
            detection_counter_label = Tk.Label( iframe, text= 'Beam Intensity Counter:' + label_spacing )
            detection_counter_label.grid( row=grid_row, column=0, sticky=Tk.E )
            detection_counter_cbox = ttk.Combobox( iframe, textvariable=self.detection_counter, width=text_entry_width, justify=Tk.CENTER )
            values = [ 'None' ]
            # TODO: make sure that self.pilatus_counter is not None
            #       and handle the situation where 'detection_counter' is not available
            if self.pilatus_counter:
                for counter in self.pilatus_counter.available_counters():
                    values.append( counter )
            detection_counter_cbox['values'] = values
            detection_counter_cbox.grid( row=grid_row, column=1, sticky=Tk.W )

            grid_row += 1
            self.postfix_syn = Tk.StringVar()
            self.postfix_syn.set( get_preference( 'postfix_syn' ) )
            postfix_syn_label = Tk.Label( iframe, text= 'Synthesized File Postfix:' + label_spacing )
            postfix_syn_label.grid( row=grid_row, column=0, sticky=Tk.E )
            postfix_syn_entry = Tk.Entry( iframe, textvariable=self.postfix_syn, width=text_entry_width, justify=Tk.CENTER )
            postfix_syn_entry.grid( row=grid_row, column=1, sticky=Tk.W )

        if self.action in [ 0, 1, 2 ]:

            grid_row += 1
            color_map_label = Tk.Label( iframe, text= 'Color Mapping Scheme:' + label_spacing )
            color_map_label.grid( row=grid_row, column=0, sticky=Tk.E )

            self.color_map = Tk.StringVar()
            self.color_map.set( get_preference( 'color_map' ) )
            j = 0
            for t, n in ( [ 'ALBULA Like', 'ALBULA' ], [ 'Diverging', 'Diverging' ] ):
                b = Tk.Radiobutton( iframe, text=t,
                                variable=self.color_map, value=n )
                j += 1
                b.grid( row=grid_row, column=j, sticky=Tk.W )

        if self.action in [ 3 ]:
            grid_row += 1
            syn_images_label = Tk.Label( iframe, text= 'Image Files to Synthesize:' + label_spacing )
            syn_images_label.grid( row=grid_row, column=0, sticky=Tk.E )

            self.syn_policy = Tk.StringVar()
            self.syn_policy.set( get_preference( 'syn_policy' ) )
            j = 0
            for t, policy in ( [ 'All', 'all' ], [ 'Select ⇒', 'select']  ):
                b = Tk.Radiobutton( iframe, text=t,
                                variable=self.syn_policy, value=policy )
                b.grid( row=grid_row+j, column=1, sticky=Tk.W )
                j += 1

            if self.syn_policy.get() == 'all':
                state = 'disabled'
            else:
                state = 'normal'

            grid_row += 1
            flags = get_preference( 'syn_flags' )
            self.image_syn_flags = []
            cbs = []
            j = 0
            for text in ( '*_0_*.tif', '*_1_*.tif', '*_2_*.tif' ):
                ivar = Tk.IntVar()
                ivar.set( flags[j] )
                self.image_syn_flags.append( ivar )
                state_ = state
                if j == 0:
                    state_ = 'disabled'         # 当面原画は必須（変更不可）とする。
                b = Tk.Checkbutton( iframe, text='%-22.22s' % (text), variable=ivar, state=state_ )
                cbs.append( b )
                b.grid( row=grid_row, column=2+j, sticky=Tk.W )
                j += 1

            def syn_policy_trace( *args ):
                policy = self.syn_policy.get()
                if policy == 'all':
                    state = 'disabled'
                else:
                    state = 'normal'
                j = 0
                for b in cbs:
                    if policy == 'all':
                        self.image_syn_flags[j].set( 1 )
                    state_ = state
                    if j == 0:
                        state_ = 'disabled'     # 当面原画は必須（変更不可）とする。
                    b.config( state=state_ )
                    j += 1

            self.syn_policy.trace( "w", syn_policy_trace )

            self.trace_  = True

            def syn_select_image_trace( *args ):
                if not self.trace_:
                    # 以下のチェック後の処置中の呼び出しを無視する。
                    # なぜか UnboundLocalError: local variable 'trace_' referenced before assignment
                    # のためメンバー変数とした。
                    return

                count = 0
                for ivar in self.image_syn_flags:
                    if ivar.get() == 1:
                        count += 1
                if count < 2:
                    MessageBox.showinfo( 'Select More', 'You must select at least two.' )
                    self.trace_ = False
                    self.syn_policy.set( 'all' )
                    self.trace_ = True

            for ivar in self.image_syn_flags:
                ivar.trace( "w", syn_select_image_trace )

        if self.action in [ 0 ]:
            grid_row += 1
            save_policy_label = Tk.Label( iframe, text= 'Foler/File Entries Save Policy:' + label_spacing )
            save_policy_label.grid( row=grid_row, column=0, sticky=Tk.E )
            self.save_policy = Tk.StringVar()
            self.save_policy.set( get_preference( 'save_policy' ) )
            j = 0
            for t, n in ( [ 'Ask', 'Ask' ], [ 'Save', 'Yes' ], [ 'No Save', 'No' ] ):
                b = Tk.Radiobutton( iframe, text=t,
                                variable=self.save_policy, value=n )
                j += 1
                b.grid( row=grid_row, column=j, sticky=Tk.W )

        # global grab cannot be set befor windows is 'viewable'
        # and this happen in mainloop after this function returns
        # Thus, it is needed to delay grab setting of an interval
        # long enough to make sure that the window has been made
        # 'viewable'
        if self.grab == 'global':
            self.after(100, self.grab_set_global )
        else:
            pass # local grab is set by parent class constructor

    def apply( self ):  # overrides parent class method
        print( "ok. apply" )

        if self.action in [ 0, 3 ]:
            set_preference( 'syn_method', self.syn_method.get() )
            old_counter_id = get_preference( 'detection_counter' )
            new_counter_id = self.detection_counter.get()
            if new_counter_id != old_counter_id:
                self.couter_id_changed = True
            set_preference( 'detection_counter', new_counter_id )
            set_preference( 'postfix_syn', self.postfix_syn.get() )

        if self.action in [ 0, 1, 2 ]:
            set_preference( 'color_map', self.color_map.get() )

        if self.action in [ 3 ]:
            flags = []
            for ivar in self.image_syn_flags:
                flags.append( ivar.get() )
            # print( 'self.image_syn_flags=', flags )
            set_preference( 'syn_flags', flags )

        if self.action in [ 0 ]:
            set_preference( 'save_policy', self.save_policy.get() )
            save_preferences()

        self.applied    = True



'''
=====================================
    Rotating 3D voxel animation
=====================================

from https://gist.github.com/mtesseracted/62ba9204d5fd3bb952c830c923c1b321

Demonstrates using ``ax.voxels`` with uneven coordinates
'''
import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.animation as manimation
from math import copysign

# Define the letters in a doc string, hashes will be 
#  colored in and the rotation order is alphabetical,
#  starting at A. The letters on the corners show
#  where the top left corner will be when viewing
#  that face.
faceStr="""
     ______
     |#   |
     |####|
     |####|
     C#   |
A====B====F====D=====
|####|#  #|#  #|#  #|
|#  #|#  #|## #|#  #|
|####|####|####|####|
|#   |   #|# ##|#  #|
|#   |####|#  #|#  #|
=====================
          | ## |
          |#  #|
          |#  #|
          | ## |
          E=====
"""

# Box colors
boxcol  = '#909cd540'
edgecol = '#7D84A6'
# 739ec040
# 7a88cc
# Set the color of each face in RGBA hex, in rotation order
fcol1 = '#ffd030d0'
faceCols=[fcol1]*6 # set them all the same


def explode(data):
    '''
    Make a new grid where the original grid is every other point
    in the new grid, starting at the third point.  Also add an
    extra padding row/col when a dimension is even to prevent
    gaps in the middle when viewing that dimension face on.

    Parameters
    ----------
    data: 3D np.array of coordinates
    '''

    size = np.array(data.shape)
    dmxx, dmxy, dmxz = (size * 2) + 1  # max dimension values

    # add extra row/col if even and then double
    size = np.array([x+1 if (x % 2) == 0 else x for x in size])*2

    # make and fill the exploded data
    data_e = np.zeros(size + 1, dtype=data.dtype)
    data_e[2:dmxx:2, 2:dmxy:2, 2:dmxz:2] = data

    return data_e


def voxel_face(corns, dm, nf, offst=0.0):
    '''
    Grab the corner coordinates of one voxel face

    Parameters
    ----------
    corns : np.indices array of corners for one voxel
    dm : (dimension), values can be  0(x), 1(y), 2(z)
    nf : (near/far face), values can be 0(near), 1(far)
    offst: how much to offset the face size, also offsets at
           1/10th that amount in the perpendicular direction
    '''

    lc = corns.copy()  # local copy so we don't swap original
    if dm == 1:  # swap y into x and correct ordering
        lc[0], lc[1] = corns[1].transpose(1, 0, 2), corns[0].transpose(1, 0, 2)
    if dm == 2:  # swap z into x and correct ordering
        lc[0], lc[2] = corns[2].transpose(2, 1, 0), corns[0].transpose(2, 1, 0)

    ret = np.zeros((3, 2, 2))
    xc1 = lc[0, nf, 0, 0]  # hold x dim constant
    xc1 += offst * (2*nf - 1) / 10.0  # offset a little
    ret[0, :] = np.array([[xc1, xc1], [xc1, xc1]])
    yc1, yc2 = lc[1, 0, 0:2, 0]
    yc1 += offst  # shrink to show voxel edges
    yc2 -= offst
    ret[1, :] = np.array([[yc1, yc2], [yc1, yc2]])
    zc1, zc2 = lc[2, 0, 0, 0:2]
    zc1 += offst  # shrink to show voxel edges
    zc2 -= offst
    ret[2, :] = np.array([[zc1, zc1], [zc2, zc2]])

    if dm != 0:  # swap x back into desired dimension
        ret[0], ret[dm] = ret[dm].copy(), ret[0].copy()
    return ret


def create_face(voxels, side, facestr):
    '''
    Change voxels to True for cubes corresponding to
    '#' symbols in facestr on the side specified

    Parameters
    ----------
    voxels: 3D array of bools
    side: string representing the constant face
    facestr: string with '#' representing true and
             rows seperated by newlines
    '''

    sides = {
        #    (start, for_x, for_y)
        'A': ([0, 0, -1], [1, 0, 0], [0, 0, -1]),    # y0
        'B': ([-1, 0, -1], [0, 1, 0], [0, 0, -1]),   # x1
        'C': ([0, 0, -1], [0, 1, 0], [1, 0, 0]),   # z1
        'D': ([0, -1, -1], [0, -1, 0], [0, 0, -1]),  # x0
        'E': ([0, 0, 0], [1, 0, 0], [0, 1, 0]),   # z0
        'F': ([-1, -1, -1], [-1, 0, 0], [0, 0, -1]), # y1
    }
    #'B': ([-1, 0, -1], [0, 1, 0], [0, 0, -1]),   # x1
    #'B': ([-1, -1, -1], [0, -1, 0], [0, 0, -1]),   # x1
    start, for_x, for_y = np.array(sides.get(side))

    for out_y, line in enumerate(facestr.strip('\r\n').splitlines()):
        #for out_x, char in enumerate(line.strip()):
        for out_x, char in enumerate(line):

            x, y, z = start + for_x*out_x + for_y*out_y
            voxels[x, y, z] = (char == '#')

    return



sarr = faceStr.strip('\r\n').split('\n')
lins = [None]*len(sarr)
for i in range(len(sarr)):
    lins[i] = sarr[i].strip(' ')
    #np.fromiter(sarr[i].strip(' '), dtype='S1')

iinds = np.zeros((6,2), dtype=int) #upper and lower lims, alphabetical
jinds = np.zeros((6,2), dtype=int) #vertical lims

# Find the index bounds from faceStr, hard-coded logic
# Key:: A:0, B:1, C:2, D:3, E:4, F:5
#C
iinds[2, 0] = jinds[2, 0] = 0 #C is top corner
iinds[2, 1] = len(lins[0]) -1 #C width is the first line
jinds[2, 1] = [x[0] for x in lins].index('A') #goes until A
#A
iinds[0, 0] = 0 #A is far left
jinds[0, 0] = jinds[2,1] #A_j starts at C end
jinds[0, 1] = [x[0] for x in lins].index('=') #A_j goes until E
iinds[0, 1] = lins[jinds[0, 0]].index('B') #A goes until B
#B
iinds[1, 0] = iinds[0, 1] #B_i starts at A end
iinds[1, 1] = iinds[1, 0] + (iinds[2, 1] - iinds[2, 0]) #same width as C
jinds[1, :] = jinds[0, :] #B_j same as A
#F
iinds[5, 0] = iinds[1, 1] 
iinds[5, 1] = iinds[5, 0] + (iinds[0, 1] - iinds[0, 0]) #same width as A
jinds[5, :] = jinds[0, :] #F_j same as A
#D
iinds[3, 0] = iinds[5, 1] 
iinds[3, 1] = iinds[3, 0] + (iinds[2, 1] - iinds[2, 0]) #same width as C
jinds[3, :] = jinds[0, :] #D_j same as A

#E
iinds[4, :] = iinds[0, :] #E_i same as F
jinds[4, 0] = jinds[0, 1] #top E == bottom A
jinds[4, 1] = jinds[4, 0] + (iinds[2, 1] - iinds[2, 0] ) #Same height as C

jinds[:, 1] -= 1 # shave off bottom '='
#'''

# build letters, @TODO: change hard-code dims
#n_voxels = np.zeros((4, 4, 5), dtype=bool)
ylen = iinds[2,1] - iinds[2,0] - 1
xlen = iinds[0,1] - iinds[0,0] - 1
zlen = jinds[0,1] - jinds[0,0] 
n_voxels = np.zeros((xlen, ylen, zlen), dtype=bool)

letters = [None]*6
lett_face = np.zeros((6, 2), dtype=int)

jinds[:, 1] -= 1 #shave off bottom '='
for i in range(6): #len(sarr)):
    i1, i2 = iinds[i, :]
    j1, j2 = jinds[i, :]
    istr = "\n"
    for j in range(j2-j1+1):
        istr += lins[j1+j+1][i1+1:i2] + '\n'
        #print(lins[j1+j+1][i1+1:i2])

    create_face(n_voxels, chr(ord('A') + i), istr)
    letters[i] = np.array(np.where(n_voxels)).T
    n_voxels[...] = False

lett_face[0] = [1, 0]  # close y face
lett_face[1] = [0, 1]  # far x face 
lett_face[2] = [2, 1]  # far z face
lett_face[3] = [0, 0]  # close x face
lett_face[4] = [2, 0]  # close z face
lett_face[5] = [1, 1]  # far y face

fcol = np.full(n_voxels.shape, boxcol)
ecol = np.full(n_voxels.shape,  edgecol)
filled = np.ones(n_voxels.shape)

# upscale the above voxel image, leaving gaps
filled_2 = explode(filled)
fcolors_2 = explode(fcol)
ecolors_2 = explode(ecol)

# Shrink the gaps
corn = np.indices(np.array(filled_2.shape) + 1).astype(float) // 2
ccorn = 0.05  # close corner
fcorn = 1.0 - ccorn
corn[0, 0::2, :, :] += ccorn
corn[1, :, 0::2, :] += ccorn
corn[2, :, :, 0::2] += ccorn
corn[0, 1::2, :, :] += fcorn
corn[1, :, 1::2, :] += fcorn
corn[2, :, :, 1::2] += fcorn
#corn[...] += 2.5
#corn[...] *= 1000.0
# print(corn[:,:2,:2,:2])
#print(corn.shape)

fig = plt.figure()
ax = fig.add_subplot(111, projection='3d')
ax.axis("off")

# Plot the voxels
x, y, z = corn
ax.voxels(x, y, z, filled_2, facecolors=fcolors_2, edgecolors=ecolors_2)

# Plot the letter square faces
jj = 0
myo = 0.005  # my face offset
for j in [x for x in letters if x is not None]:

    locf = np.empty((j.shape[0], 3, 2, 2))  # local face

    ji = 0
    for i in j:
        i = (i + 1) * 2  # skip empty voxels
        loc = corn[:, i[0]:i[0]+2, i[1]:i[1]+2, i[2]:i[2]+2]  # local corners
        locf[ji] = voxel_face(loc, lett_face[jj, 0], lett_face[jj, 1], myo)
        ax.plot_surface(locf[ji, 0], locf[ji, 1], locf[ji, 2], 
                        color=faceCols[jj], shade=False)
        # ffe749
        # ffe500
        ji += 1

    jj += 1

# plt.show()
#'''
ssp =  8 #small step
bsp = 40 #big step
sps = 15 #short pause
#Views:        PY,  P,   Y,  T,   H,   O,    N,   spiiiiiiiiiiiiin,  PY
view_elev = [  5,   0,   0,  90,   0, -90,   0,   5,  10,  11,  10,   5]
view_azim = [-60, -90,   0,  90, 180, 180,  90,  60,  30,   0, -30, -60]
view_step = [ssp, bsp, bsp, bsp, bsp, bsp, ssp, ssp, ssp, ssp, ssp]
view_paus = [ 40, sps, sps, sps, sps, sps, sps,   0,   0,   0,   0]

FFMpegWriter = manimation.writers['ffmpeg']
metadata = dict(title='Movie Test', artist='Matplotlib',
                comment='Movie support!')
writer = FFMpegWriter(fps=25, metadata=metadata)

with writer.saving(fig, "voxelRotate.mp4", 200):


    for i in range(1,len(view_elev)):

        de = (view_elev[i] - view_elev[i-1])
        da = (view_azim[i] - view_azim[i-1])

        if abs(da) >= 180 : #unecessary in this config
            da -= copysign(360, da)
        if abs(de) >= 180 :
            de -= copysign(360, de)

        steps = view_step[i-1]
        da = da / steps
        de = de / steps

        for j in range(view_paus[i-1]): #Pause on direct view of a letter
            ax.view_init(view_elev[i-1], view_azim[i-1])
            plt.draw()
            writer.grab_frame()
        for j in range(steps): #Rotate to next letter
            ax.view_init(view_elev[i-1] + j*de,
                         view_azim[i-1] + j*da)
            plt.draw()
            writer.grab_frame()
#'''


# coding: utf-8
"""
    OurCurvature.py

    Copyright (c) 2016-2017, Masatsuyo Takahashi, KEK-PF
"""
import numpy                as np
from scipy.interpolate      import UnivariateSpline

def curvature_curve( x, y, spline=None ):

    if spline is None:
        fy = UnivariateSpline( x, y, k=3, s=0 )
    else:
        fy = spline

    yˈ  = fy.derivative(1)(x)
    yˈˈ = fy.derivative(2)(x)

    cy = yˈˈ / np.power( 1 + yˈ**2, 3/2 )

    return cy



"""
    SecTheory.BasicModels.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.special import iv, ive
from scipy.interpolate import UnivariateSpline

def single_pore_pdf(t, np_, tp_):
    return iv(1, np.sqrt(4*np_*t/tp_)) * np.sqrt(np_/(t*tp_)) * np.exp(-t/tp_-np_)

SMALL_POS_VALUE = 1e-8

def robust_single_pore_pdf(t, np_, tp_, debug=False):
    # Bessel functions in Python that work with large exponents
    # https://stackoverflow.com/questions/13726464/bessel-functions-in-python-that-work-with-large-exponents
    #
    # iv(1, np.sqrt(4*np_*t/tp_)) * np.sqrt(np_/(t*tp_)) * np.exp(-t/tp_-np_)
    #
    # ive(v, z) = iv(v, z) * exp(-abs(z.real))
    # iv(v, sq) = ive(v, sq) * exp(sq)

    # val = single_pore_pdf(t, np_, tp_)
    sq = np.sqrt(4*np_*t/tp_)
    val = ive(1, sq) * np.sqrt(np_/(t*tp_)) * np.exp(sq -t/tp_ -np_)
    isnan_val = np.isnan(val)
    if debug:
        import molass_legacy.KekLib.DebugPlot as plt

        wh = np.where(isnan_val)[0]
        if len(wh) > 0:
            i, j = wh[[0, -1]]
            if i > 0 and val[i-1] > SMALL_POS_VALUE and j < len(val) - 1 and val[j+1] > SMALL_POS_VALUE:
                print("np_=", np_)
                print("tp_=", tp_)
                print("wh=", wh)
                if j < len(val) - 1:
                    print("val[j+1]=", val[j+1])
                isfinite = np.isfinite(val)
                with plt.Dp():
                    fig, ax = plt.subplots()
                    ax.plot(t, val)
                    ax.plot(t[isnan_val], np.zeros(len(wh)), color="red")
                    fig.tight_layout()
                    plt.show()
    val[isnan_val] = 0
    return val

def fit_single_pore(x, y, moments=None):
    from scipy.optimize import minimize

    def f(p):
        h, t0, np_, tp_ = p
        y_ = h * robust_single_pore_pdf(x - t0, np_, tp_)
        return np.sum((y_ - y)**2)

    t0 = 0
    np_ = np.mean(x) if moments is None else moments[1]
    tp_ = 1
    ret = minimize(f, (0.1, t0, np_, tp_))
    return ret.x

def single_pore_height(np_, tp_):
    return ive(1, 2*np_) * np.sqrt(1/tp_**2)

def robust_single_pore_pdf_scaled(t, np_, tp_):
    return robust_single_pore_pdf(t, np_, tp_)/single_pore_height(np_, tp_)

def single_pore_elution(x, rg, rp, nperm, tperm, me, mp):
    if rg > rp:
        np_ = 0
        tp_ = 0
    else:
        rho = rg/rp
        np_ = nperm*(1 - rho)**me
        tp_ = tperm*(1 - rho)**mp
    return robust_single_pore_pdf(x, np_, tp_)

class SimpleSec:
    def __init__(self, rp0=124.34, nperm=2000, tperm=1):
        self.rp0 = rp0
        self.nperm = nperm
        self.tperm = tperm

    def get_elution(self, x, rg):
        return single_pore_elution(x, rg, self.rp0, self.nperm, self.tperm, 1, 1)

if __name__ == '__main__':
    import os
    import sys
    lib_dir = os.path.dirname(os.path.dirname(__file__))
    sys.path.append(lib_dir)
    kek_lib = os.path.join(lib_dir, "KekLib")
    sys.path.append(kek_lib)
    import molass_legacy.KekLib.DebugPlot as plt

    t = 10 + np.arange(300)
    y = robust_single_pore_pdf(t, 150, 1, debug=True)
    with plt.Dp():
        fig, ax = plt.subplots()
        ax.plot(t, y)
        plt.show()



"""
    Models.Stochastic.MonoporeUtils.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
import molass_legacy.KekLib.DebugPlot as plt
from SecTheory.BasicModels import robust_single_pore_pdf
from molass_legacy.Models.Stochastic.MonoporeUtilsImpl import compute_monopore_curves

def plot_monopore_curves(ax, x, monopore_params, rgs, return_artists=False):
    cy_list, ty = compute_monopore_curves(x, monopore_params[0:6], rgs, monopore_params[6:])
    artists = []
    for k, cy in enumerate(cy_list):
        curve, = ax.plot(x, cy, ":", label='component-%d' % k)
        artists.append(curve)
    curve, = ax.plot(x, ty, ":", color="red", label='model total')
    artists.append(curve)
    if return_artists:
        return artists

def adjust_scales_to_egh_params_array(x, y, adjust_params, rgs, peaks, debug=False):
    N, T, x0, me, mp, poresize = adjust_params
    rhov = rgs/poresize
    rhov[rhov > 1] = 1
    scales = []
    if debug:
        cy_list = []
    for k, (rho, params) in enumerate(zip(rhov, peaks)):
        ni_ = N*(1 - rho)**me
        ti_ = T*(1 - rho)**mp
        cy = robust_single_pore_pdf(x - x0, ni_, ti_)
        if debug:
            cy_list.append(cy)
        scale = params[0]/np.max(cy)
        print([k], params, scale)
        scales.append(scale)

    if debug:
        def plot_curves(ax, cy_list, scales=None):
            ax.plot(x, y, label='data') 
            for k, cy in enumerate(cy_list):
                if scales is not None:
                    cy *= scales[k]
                ax.plot(x, cy, ":", label='component-%d' % k)
            ty = np.sum(cy_list, axis=0)
            ax.plot(x, ty, ":", color="red", label='model total')
            ax.legend()

        with plt.Dp():
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            fig.suptitle("PDF vs. Scaled", fontsize=20)
            ax1.set_title("PDF", fontsize=16)
            plot_curves(ax1, cy_list)
            ax2.set_title("Scaled", fontsize=16)
            plot_curves(ax2, cy_list, scales=scales)
            fig.tight_layout()
            plt.show()

    return scales

def draw_exclusion_cuve(axt, monopore_columnparams, trs, rgs):
    N, T, x0, me, mp, poresize = monopore_columnparams
    if np.isscalar(poresize):
        rv = np.linspace(poresize, 10, 100)
        rhov = rv/poresize
        tv = x0 + N*T*(1 - rhov)**(me+mp)
        axt.plot(tv, rv, lw=2, color="yellow")
        axt.plot(trs, rgs, "o", color="red")
        axt.axvline(x=x0, color="red")

def plot_monopore_moments_state(title, x, y, peak_rgs, mnp_params, x0_upper, cy_list, ty, moments, egh_moments_list):
    from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
    import Models.Stochastic.MonoporeUtilsImpl as module

    def plot_monopore_component_with_sliders():
        from importlib import reload
        import Models.Stochastic.MonoporeUtilsImpl as module
        reload(module)
        from molass_legacy.Models.Stochastic.MonoporeUtilsImpl import plot_monopore_component_with_sliders_impl
        plot_monopore_component_with_sliders_impl(title, x, y, peak_rgs, mnp_params, x0_upper, cy_list, ty, moments, egh_moments_list)

    print("plot_monopore_moments_state: peak_rgs=", peak_rgs)

    N, T, x0_, me, mp, poresize = mnp_params[0:6]
    num_peaks = len(peak_rgs)
    extra_button_specs = [("Plot Component with Sliders", plot_monopore_component_with_sliders)]
    with plt.Dp(button_spec=["OK", "Cancel"],
                extra_button_specs=extra_button_specs):
        fig, (ax1,ax2) = plt.subplots(ncols=2, figsize=(12, 5))
        fig.suptitle(title)
        ax2.set_title("Moments Fitting for %s" % get_in_folder())
        ax2.set_xlabel("Time (frames)")
        ax2.set_ylabel("Intensity")
        for ax in (ax1,ax2):
            ax.plot(x, y, color='orange', label='data')
            ax.plot(x, ty, ':', color="red", label='model total')
            for k, (cy, M, eghM) in enumerate(zip(cy_list, moments, egh_moments_list)):
                ax.plot(x, cy, ':', label='component-%d' % k)
                label = None if k < num_peaks - 1 else "Monopore Moments"
                ax.axvline(x=M[0], color="green", label=label)
                label = None if k < num_peaks - 1 else "EGH Moments"
                ax.axvline(x=eghM[0], ls=":", lw=3, color="cyan", label=label)
            ax.axvline(x=x0_upper, color="pink")
            axt = ax.twinx()
            axt.grid(False)
            trs = [M[0] for M in moments]
            draw_exclusion_cuve(axt, (N, T, x0_, me, mp, poresize), trs, peak_rgs)
        # xmin, xmax = ax1.get_xlim()
        # ymin, ymax = ax1.get_ylim()
        m1, m2 = moments[0]
        xmin = m1 - 10*np.sqrt(m2)
        m1, m2 = moments[-1]
        xmax = m1 + 10*np.sqrt(m2)                
        ax2.set_xlim(xmin, xmax)
        ax2.legend()
        fig.tight_layout()
        ret = plt.show()

    return ret

def plot_monopore_conformance(K, t0, poresize, plot_info, rg_info=None, confrg_info=None, plot_cuvatures=False):
    from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder

    rv = np.linspace(10, 100, 100)
    rhov = rv/poresize
    rhov[rhov > 1] = 1
    tv = t0 + K*np.power(1 - rhov, 3)
    x = plot_info.x
    y = plot_info.y
    model = plot_info.model
    cy_list = []
    for params in plot_info.peaks:
        cy = model(x, params)
        cy_list.append(cy)

    with plt.Dp():
        if plot_cuvatures:
            fig, (ax,ax2) = plt.subplots(ncols=2, figsize=(18,5))
        else:
            fig, ax = plt.subplots(figsize=(9,5))

        fig.suptitle("Exclusion Curve Estimate on %s" % get_in_folder(), fontsize=20)

        ax.set_title("Data Exclusion Points against Estimated Exclusion Curve", fontsize=16)
        ax.set_xlabel("Time (Frames)")
        ax.set_ylabel("Intensity")
        ax.plot(x, y, color="orange")
        for cy in cy_list:
            ax.plot(x, cy, ":")

        ax.axvline(x=t0, color="red", label="$t_0$")
        ax.legend(loc="upper left")

        axt = ax.twinx()
        axt.grid(False)
        axt.set_ylabel(r"$R_g (\AA_)$")
        axt.plot(tv, rv, color='yellow', lw=3, label="Estimated Exclusion Curve: K=%.4g, Rp=%.3g" % (K, poresize))
        ry = None
        if rg_info is not None:
            rgs, trs = rg_info[0:2]
            axt.plot(trs, rgs, "o")
            if len(rg_info) > 2:
                from scipy.interpolate import UnivariateSpline
                if rg_info[2]:  # draw spline
                    spline = UnivariateSpline(trs, rgs)
                ry = spline(tv)
                axt.plot(tv, ry, color="cyan", alpha=0.5, lw=3, label="Spline of Exclusion Points")
                
        if confrg_info is not None:
            rgs, trs = confrg_info
            axt.plot(trs, rgs, "o")

        ymin, ymax = axt.get_ylim()
        axt.set_ylim(max(0, ymin), ymax)

        axt.legend()

        if plot_cuvatures:
            from molass_legacy.KekLib.OurCurvature import curvature_curve
            ax2.set_title("Corresponding Curvatures", fontsize=16)
            ax2.set_xlabel("Time (Frames)")
            ax2.set_ylabel("Curvature")

            # to avoid ValueError: x must be strictly increasing if s = 0
            t = np.flip(tv)     
            diff_t = np.diff(t)
            diff_t_pos = np.where(diff_t > 0)[0]
            p = diff_t_pos[0]
            t_ = t[p:]
            try:
                cy = curvature_curve(t_, np.flip(rv)[p:])

                # to avoid anomalies
                m = np.argmax(cy)
                # there can be anomalies beyond m as in 20191006_OA_Ald_Ferr
                diff_cy = np.diff(cy[m:])
                diff_cy_pos = np.where(diff_cy > 0)[0]
                if len(diff_cy_pos) > 0:
                    m_ = m + diff_cy_pos[-1] + 1
                else:
                    # as in 20210727/data02 
                    m_ = m

                ax2.plot(t_[m_:], cy[m_:], color="yellow", lw=3, label="Estimated Exclusion Curve Curvature")
            except:
                from molass_legacy.KekLib.ExceptionTracebacker import log_exception
                log_exception(None, "curvature_curve rv failure: ")
            if ry is not None:
                try:
                    cy_ = curvature_curve(t_, np.flip(ry)[p:])
                    ax2.plot(t_, cy_, color="cyan", alpha=0.5, lw=3, label="Spline Curvature of Exclusion Points")
                except:
                    from molass_legacy.KekLib.ExceptionTracebacker import log_exception
                    log_exception(None, "curvature_curve ry failure: ")
            ax2.legend()

        fig.tight_layout()
        ret = plt.show()

    return ret

def guess_monopore_colparams(rgs, qualities, trs, init_params, debug=False, plot_info=None):

    def objective(p):
        K, t0, poresize = p
        rhov = rgs/poresize
        rhov[rhov > 1] = 1
        trs_ = t0 + K*(1 - rhov)**3
        return np.sum((trs_ - trs)**2 * qualities)

    init_params_ = init_params[1:]      # optimize except for N
    res = minimize(objective, init_params_)

    if debug and plot_info is not None:
        ret = plot_monopore_conformance(*res.x, plot_info, rg_info=(rgs, trs, True), plot_cuvatures=False)
        if not ret:
            return

    return np.concatenate([[1000], res.x])      # always return N=1000



# coding: utf-8
"""
    Pomegranate.EghDistribution.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np
from pomegranate import *
from molass_legacy.Peaks.ElutionModels import egh_pdf, compute_egh_params
import molass_legacy.KekLib.DebugPlot as plt

class EghDistribution:
    def __init__(self, tR, sigma, tau=0):
        self.name = 'EghDistribution'
        self.frozen = False
        self.tR = tR
        self.sigma = sigma
        self.tau = tau
        self.parameters = (self.tR, self.sigma, self.tau)
        self.d = 1
        self.summaries = numpy.zeros(4)

    def to_dict(self):
        return {
            'class' : 'Distribution',
            'name'  : self.name,
            'parameters' : self.parameters,
            'frozen' : self.frozen
        }

    def probability(self, X):
        X = X.reshape(X.shape[0])
        return egh_pdf(X, self.tR, self.sigma, self.tau)

    def log_probability(self, X):
        X = X.reshape(X.shape[0])
        return np.log(egh_pdf(X, self.tR, self.sigma, self.tau))

    def summarize(self, X, w=None):
        if w is None:
            w = numpy.ones(X.shape[0])

        X = X.reshape(X.shape[0])
        self.summaries[0] += w.sum()
        w_ = w/self.summaries[0]
        self.summaries[1] += X.dot(w_)
        self.summaries[2] += ((X - self.summaries[1])** 2).dot(w_)
        self.summaries[3] += ((X - self.summaries[1])** 3).dot(w_)

    def from_summaries(self, inertia=0.0):
        init_params = (self.summaries[1], np.sqrt(self.summaries[2]), 0)
        self.tR, self.sigma, tau = compute_egh_params(init_params, self.summaries[1:])
        moderate_tau = min(self.sigma, max(-self.sigma, tau))
        self.tau = moderate_tau
        self.parameters = (self.tR, self.sigma, moderate_tau)
        self.clear_summaries()

    def clear_summaries(self, inertia=0.0):
        self.summaries = numpy.zeros(4)

    @classmethod
    def from_samples(cls, X, weights=None):
        d = EghDistribution(0, 1)
        d.summarize(X, weights)
        d.from_summaries()
        return d

    @classmethod
    def blank(cls):
        return EghDistribution(0, 1)



"""
    PreviewButtonFrame.py

    Copyright (c) 2018-2025, SAXS Team, KEK-PF
"""
import sys
import logging
from molass_legacy.KekLib.OurTkinter import Tk
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting
from .PreviewData import PreviewData, PreviewOptions
from molass_legacy.KekLib.BasicUtils import Struct
from molass_legacy.LRF.PnoScdMap import PnoScdMap
DEBUG_SD = False
if DEBUG_SD:
    from molass_legacy.SerialAnalyzer.SdDebugger import SdDebugger

ADDITIONAL_CONSTRAINTS = False
DEVELOP_MODE = sys.executable.find("pythonw") < 0
# DEVELOP_MODE = False
ELUTION_DATATYPE_NAMES = ["Xray elution model", "Xray measured data", "UV elution model", "mapped UV data"]

class PreviewButtonFrame(Tk.LabelFrame):
    def __init__(self, parent, *args, **kwargs):
        self.logger = logging.getLogger(__name__)
        self.dialog = kwargs.pop('dialog', None)
        self.use_elution_models =  kwargs.pop('use_elution_models', None)
        self.judge_holder =  kwargs.pop('judge_holder', None)
        conc_opts = kwargs.pop('conc_opts', False)
        self.enable_conc_opts = conc_opts and get_setting('enable_conc_opts')
        self.enable_conctype_change =  get_setting('enable_conctype_change')
        self.editor =  kwargs.pop('editor', None)
        self.pre_proc =  kwargs.pop('pre_proc', None)
        self.post_proc =  kwargs.pop('post_proc', None)
        self.layout_check_only = kwargs.pop('layout_check_only', False)
        self.mapper = self.dialog.mapper
        self.is_microfluidic = self.mapper is None
        self.allow_rank_variation = get_setting('allow_rank_variation')
        self.extended_conc_dep = get_setting('extended_conc_dep')

        Tk.LabelFrame.__init__(self, parent, text="Preview Options", labelanchor=Tk.N, *args, **kwargs)

        # concentration_datatype
        conc_datatype = get_setting("concentration_datatype")
        self.concentration_datatype = Tk.IntVar()
        if self.use_elution_models:
            rb_dtypes = [0, 1, 2, 3]
            if conc_datatype is None:
                datatype = 2
            else:
                if self.enable_conctype_change:
                    datatype = conc_datatype
                else:
                    datatype = 2
        else:
            rb_dtypes = [1, 3]
            if conc_datatype is None:
                datatype = 3
            else:
                if self.enable_conctype_change:
                    if conc_datatype in rb_dtypes:
                        datatype = conc_datatype
                    else:
                        datatype = 1 if conc_datatype < 2 else 3
                else:
                    datatype = 3
        self.concentration_datatype.set(datatype)

        if self.layout_check_only:
            self.main_peak_rank = 1
        else:
            cnv_ranges = self.get_cnv_ranges()
            self.pno_map = PnoScdMap(self.dialog.sd, cnv_ranges)
            self.main_peak_rank = self.pno_map.get_main_peak_rank()

        # 
        iframe = Tk.Frame(self)
        iframe.pack(padx=0, pady=10)

        self.widgets = []

        grid_row = 0
        grid_col = 0

        self.rank_control = Tk.IntVar()
        rank_control = get_setting( 'rank_control' )
        self.rank_control.set(rank_control)
        cb = Tk.Checkbutton(iframe, text="Automatic Rank Control", variable=self.rank_control )
        cb.grid( row=grid_row, column=0, sticky=Tk.W, padx=10 )

        self.rc_dependents = []
        manual_state = Tk.DISABLED if rank_control else Tk.NORMAL

        grid_row += 1
        self.conc_dependence = Tk.IntVar()
        self.set_conc_depend()
        dependence_frame = Tk.Frame(iframe)
        label = Tk.Label(dependence_frame, text="Conc. Dependency", state=manual_state)
        label.pack(side=Tk.LEFT)
        self.rc_dependents.append(label)
        max_dependence = 6 if self.extended_conc_dep else 2
        spinbox = Tk.Spinbox( dependence_frame, textvariable=self.conc_dependence,
                                from_=1, to=max_dependence, increment=1,
                                justify=Tk.CENTER, width=6, state=manual_state)

        spinbox.pack(side=Tk.LEFT)
        self.rc_dependents.append(spinbox)
        dependence_frame.grid(row=grid_row, column=grid_col, sticky=Tk.W, padx=10 )
        self.widgets.append(spinbox)

        if self.is_microfluidic:
            grid_col += 1
            self.denat_dependent = Tk.IntVar()
            self.denat_dependent.set(get_setting('denat_dependent'))
            cb = Tk.Checkbutton(iframe, text="denat. dependent", variable=self.denat_dependent, state=manual_state)
            cb.grid( row=grid_row, column=grid_col, sticky=Tk.W, padx=10 )
            self.rc_dependents.append(cb)
            self.denat_dependent_cb = cb
            self.widgets.append(cb)

        if self.enable_conc_opts:
            grid_col += 1
            self.conc_opts_button = btn = Tk.Button( iframe, text='Conc. Options', command=self.show_conc_opts_dialog)
            btn.grid( row=grid_row, column=grid_col, sticky=Tk.W, padx=20 )
            self.widgets.append(btn)

        grid_col += 1
        rowspan_ = 7 if self.allow_rank_variation else 6
        self.show_zx_preview_button = btn = Tk.Button( iframe, text='Preview', command=self.show_zx_preview )
        btn.grid( row=0, rowspan=rowspan_, column=2, padx=30, sticky=Tk.E )
        self.widgets.append(btn)

        if False:
            self.test_button = btn = Tk.Button(iframe, text='Test', command=self.do_devel_test)
            btn.grid(row=0, rowspan=rowspan_, column=3, padx=10)

        grid_row += 1
        self.svd_reconstruct_dependents = []
        if self.allow_rank_variation:
            svd_frame = Tk.Frame(iframe)
            svd_frame.grid( row=grid_row, column=0, columnspan=2, sticky=Tk.W)
            self.svd_reconstruct = Tk.IntVar()
            self.svd_reconstruct.set( get_setting( 'svd_reconstruct' ) )
            cb = Tk.Checkbutton(svd_frame, text="SVD Reconstruction", variable=self.svd_reconstruct, state=manual_state )
            cb.grid( row=0, column=0, sticky=Tk.W, padx=10 )
            self.rc_dependents.append(cb)
            self.widgets.append(cb)

            self.rank_increment = Tk.IntVar()
            self.rank_increment.set( get_setting( 'rank_increment' ) )
            additional_rank_frame = Tk.Frame(svd_frame)
            label = Tk.Label(additional_rank_frame, text="Rank Increment", state=manual_state)
            label.pack(side=Tk.LEFT)
            self.rc_dependents.append(label)
            self.svd_reconstruct_dependents.append(label)
            self.ar_spinbox = spinbox = Tk.Spinbox(additional_rank_frame, textvariable=self.rank_increment,
                                    from_=0, to=5, increment=1,
                                    justify=Tk.CENTER, width=6, state=manual_state )

            spinbox.pack(side=Tk.LEFT)
            self.rc_dependents.append(spinbox)
            additional_rank_frame.grid(row=0, column=1, sticky=Tk.W, padx=10 )
            self.svd_reconstruct_dependents.append(spinbox)

        enable_new_features = get_setting('enable_new_features')

        # LRF Bound Correction
        self.lrf_bound_correction = Tk.IntVar()
        self.lrf_bound_correction.set(get_setting('lrf_bound_correction'))
        if enable_new_features:
            grid_row += 1
            cb = Tk.Checkbutton( iframe, text="LRF Bound Correction", variable=self.lrf_bound_correction )
            cb.grid( row=grid_row, column=0, sticky=Tk.W, padx=10 )
            self.lrf_bound_correction_cb = cb

        grid_row += 1
        self.force_proportions = Tk.IntVar()
        self.proportions_text = Tk.StringVar()
        self.proportions_text.set("0.2, 0.8")

        proportions_frame = Tk.Frame(iframe)
        proportions_frame.grid(row=grid_row, column=0, columnspan=2, sticky=Tk.W)
        state = Tk.DISABLED
        cb = Tk.Checkbutton(proportions_frame, text="Force Proportions to", variable=self.force_proportions, state=state)
        cb.grid(row=0, column=0, sticky=Tk.W, padx=10)
        self.proportions_entry = Tk.Entry(proportions_frame, textvariable=self.proportions_text, width=10, justify=Tk.CENTER, state=state)
        self.proportions_entry.grid(row=0, column=1, sticky=Tk.W, padx=10)

        # Data Selection Buttons
        # note that concentration_datatype has been created earlier above

        if self.enable_conctype_change:
            grid_row += 1
            label = Tk.Label(iframe, text="Concetration Datatype Selection:")
            label.grid(row=grid_row, column=0, padx=10, sticky=Tk.W)

            grid_row += 1
            frame = Tk.Frame(iframe)
            frame.grid(row=grid_row, column=0, columnspan=4, padx=20, sticky=Tk.W)

            for k, value in enumerate(rb_dtypes):
                text = ELUTION_DATATYPE_NAMES[value]
                rb = Tk.Radiobutton(frame, variable=self.concentration_datatype, text=text, value=value)
                d, r = divmod(k, 2)
                rb.grid(row=d, column=r, padx=10, sticky=Tk.W)

        if ADDITIONAL_CONSTRAINTS:
            grid_row += 1
            grid_col = 0
            self.aq_smoothness = Tk.IntVar()
            # this value is set in self.update_aq_smoothness
            self.aq_smoothness.set( get_setting( 'aq_smoothness' ) )
            cb = Tk.Checkbutton( iframe, text="Smoothness", variable=self.aq_smoothness, state=manual_state )
            cb.grid( row=grid_row, column=grid_col, sticky=Tk.W, padx=10 )
            self.rc_dependents.append(cb)
            self.widgets.append(cb)

            grid_col += 1
            self.aq_positivity = Tk.IntVar()
            self.aq_positivity.set( get_setting( 'aq_positivity' ) )
            cb = Tk.Checkbutton( iframe, text="A(q) Positivity", variable=self.aq_positivity, state=manual_state )
            cb.grid( row=grid_row, column=grid_col, sticky=Tk.W, padx=20 )
            self.rc_dependents.append(cb)
            self.widgets.append(cb)

        self.rank_control_tracer()
        self.svd_reconstruct.trace("w", self.rank_control_tracer)
        self.rank_control.trace("w", self.rank_control_tracer)

    def update_modelframe(self):
        self.editor = self.dialog.get_current_frame()

    def set_conc_depend(self, revive_auto_rank=False):
        if self.is_microfluidic:
            cd = 1
        else:
            cd = get_setting('conc_dependence')
            if cd is None or revive_auto_rank:
                cd = self.main_peak_rank
        self.conc_dependence.set(cd)

    def rank_control_tracer(self, *args):
        rank_control = self.rank_control.get()
        state = Tk.DISABLED if rank_control else Tk.NORMAL
        for w in self.rc_dependents:
            w.config(state=state)
        state = Tk.NORMAL if not rank_control and self.svd_reconstruct.get() else Tk.DISABLED
        for w in self.svd_reconstruct_dependents:
            w.config(state=state)
        if rank_control:
            # reset to default which has been established before the init of this frame
            self.set_conc_depend(revive_auto_rank=True)

    def get_cnv_ranges(self):
        # task: reduce redundancy with get_preview_data
        from molass_legacy.DataStructure.AnalysisRangeInfo import convert_to_paired_ranges

        paired_ranges = self.dialog.make_range_info(self.concentration_datatype.get())
        ret = convert_to_paired_ranges(paired_ranges)
        return ret[0]

    def get_preview_data(self, with_update=True):
        use_mtd_conc = get_setting('use_mtd_conc')

        if with_update:
            self.update_settings(update_memorized=False)    # False to avoid recursive calls to update_settings

        if use_mtd_conc:
            decomp_info = self.dialog.get_decomp_info()
            pdata = PreviewData(xdata=self.dialog.xdata, decomp_info=decomp_info, slice_=self.dialog.slice_)
        else:
            mapper = self.mapper
            sd_ = self.get_desired_serial_data(mapper)    # 
            # sd_ = self.dialog.sd
            paired_ranges = self.dialog.make_range_info(self.concentration_datatype.get())
            pdata = PreviewData(sd=sd_, mapper=mapper, judge_holder=self.judge_holder, paired_ranges=paired_ranges)

        """
        TODO: other params should be handled in the same way
        """
        if ADDITIONAL_CONSTRAINTS:
            aq_smoothness = self.aq_smoothness.get() == 1
            aq_positivity = self.aq_positivity.get() == 1
        else:
            aq_smoothness = False
            aq_positivity = False

        concentration_datatype = self.concentration_datatype.get()
        use_elution_models = concentration_datatype in [0, 2]
        conc_depend = self.get_conc_depend()
        popts = PreviewOptions(conc_depend, aq_smoothness, aq_positivity, use_elution_models)

        return pdata, popts

    def memorize_preview_data(self):
        pdata, popts = self.get_preview_data(with_update=False)
        set_setting('preview_params', (pdata, popts))

    def show_zx_preview(self, use_pool=True):
        if self.pre_proc is not None:
            self.pre_proc()

        self.update_settings()
        pdata, popts = get_setting('preview_params')

        if use_pool:
            self.show_zx_preview_using_pool(pdata, popts)
        else:
            from PreviewController import PreviewController
            self.preview_ctl = PreviewController(dialog=self.dialog, editor=self.editor)
            self.preview_ctl.run_solver(self.dialog, pdata, popts)
            if self.preview_ctl.ok():
                self.preview_ctl.show_dialog()
                self.update_widgets()
                if self.preview_ctl.dialog.applied:
                    self.update_settings()
            else:
                # TODO
                pass

        if self.post_proc is not None:
            self.post_proc()

    def get_last_change_id_from_advanced_frame(self):
        return self.dialog.advanced_frame.get_last_change_id()

    def show_zx_preview_using_pool(self, pdata, popts, debug=False):
        if debug:
            from importlib import reload
            import molass_legacy.LRF.LrfResultPool
            reload(molass_legacy.LRF.LrfResultPool)       
        from molass_legacy.LRF.LrfResultPool import LrfResultPool
        self.pool = LrfResultPool(pdata, popts)
        self.pool.run_solver(self.dialog)
        last_change_id = self.get_last_change_id_from_advanced_frame()
        self.pool.show_dialog(self.dialog.parent, last_change_id=last_change_id)

    def get_desired_serial_data(self, mapper):
        try:
            # try: to avoid errors like
            # _tkinter.TclError: invalid command name ".!decompeditordialog9.!frame2.!frame2.!previewbuttonframe"
            self.config(cursor='wait')
            self.update()
        except:
            pass

        sd = self.dialog.sd
        sd_copy = sd.get_exec_copy( mapper )

        if DEBUG_SD:
            debugger =SdDebugger()
            debugger.save_info(sd_copy)
        mapped_info = mapper.get_mapped_info()
        if mapped_info.needs_sd_corrcetion():
            sd_copy.apply_baseline_correction(mapped_info)
            if DEBUG_SD:
                debugger.save_info(sd_copy)
        sd_copy.absorbance = sd.absorbance    # exec_copy does not have absorbance

        try:
            self.config(cursor='')
        except:
            pass

        return sd_copy

    def config_states_alternative( self, state ):
        ignore_bq_list = get_setting('ignore_bq_list')
        for k, w in enumerate(self.widgets):
            if k == 0 and ignore_bq_list is not None:
                state_ = Tk.DISABLED
            else:
                state_ = state
            w.config( state=state_ )

    def config_states( self, state ):
        ignore_bq_list = get_setting('ignore_bq_list')
        for k, w in enumerate(self.widgets):
            w.config( state=state )

    def get_conc_depend(self):
        # 
        return self.conc_dependence.get()

    def update_widgets( self ):
        # TODO: 

        if ADDITIONAL_CONSTRAINTS:
            self.aq_smoothness.set( get_setting( 'aq_smoothness' ) )
            self.aq_positivity.set( get_setting( 'aq_positivity' ) )
        self.conc_dependence.set( get_setting( 'conc_dependence' ) )
        self.update()

    def update_settings(self, update_memorized=True):

        if hasattr(self.editor, "model_name"):
            preview_model = self.editor.model_name
        else:
            preview_model = None
        set_setting('preview_model', preview_model)

        rank_control = self.rank_control.get()
        set_setting('rank_control', rank_control)

        if ADDITIONAL_CONSTRAINTS:
            set_setting('aq_smoothness', self.aq_smoothness.get())
            set_setting('aq_positivity', self.aq_positivity.get())

        conc_dependence = self.conc_dependence.get()
        set_setting('conc_dependence', conc_dependence )
        self.logger.info('conc_dependence: %d' % conc_dependence)

        if self.is_microfluidic:
            denat_dependent = self.denat_dependent.get()
            set_setting('denat_dependent', denat_dependent)
            ignore_all_bqs = 1 - denat_dependent
        else:
            if rank_control:
                ignore_all_bqs = 0
            else:
                ignore_all_bqs = 1 if conc_dependence == 1 else 0
        set_setting('ignore_all_bqs', ignore_all_bqs)

        if self.allow_rank_variation:
            set_setting('svd_reconstruct',  self.svd_reconstruct.get())
            set_setting('rank_increment',   self.rank_increment.get())

        set_setting('lrf_bound_correction', self.lrf_bound_correction.get())
        set_setting('concentration_datatype', self.concentration_datatype.get())

        if update_memorized:
            self.memorize_preview_data()

    def show_conc_opts_dialog(self):
        from ConcOptsDialog import ConcOptsDialog
        pranges = self.editor.make_restorable_ranges()
        dialog = ConcOptsDialog(self.dialog, self.dialog, pranges)
        dialog.show()

    def set_guard_for_three_state_model(self):
        self.denat_dependent.set(0)
        self.denat_dependent_cb.config(state=Tk.DISABLED)

    def remove_guard_for_three_state_model(self):
        self.denat_dependent.set(1)
        self.denat_dependent_cb.config(state=Tk.NORMAL)

    def do_devel_test(self):
        from importlib import reload
        import Tools.EmbedCushion
        reload(Tools.EmbedCushion)
        from molass_legacy.Tools.EmbedCushion import embed_cushion

        embed_cushion(self)



"""
    MappingUtils.py

    Copyright (c) 2021-2023, SAXS Team, KEK-PF
"""

import molass_legacy.KekLib.DebugPlot as plt

def debug_plot_mapping(mapper, title="debug_plot_mapping"):
    from scipy.interpolate import UnivariateSpline

    a_curve = mapper.a_curve
    a_base = mapper.a_base
    x_curve = mapper.x_curve
    x_base = mapper.x_base

    with plt.Dp():
        fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(21, 7))
        fig.suptitle(title, fontsize=20)
        ax1.set_title("UV Elution", fontsize=16)
        ax2.set_title("Xray Elution", fontsize=16)
        ax3.set_title("Mapped Elutions", fontsize=16)

        for ax, curve, color, base in [(ax1, a_curve, None, a_base), (ax2, x_curve, 'orange', x_base)]:
            x = curve.x
            y = curve.y
            yb = y + base
            ax.plot(x, yb, color=color, label='data')
            ax.plot(x, base, ':', color='red')

            top = curve.peak_top_x
            bnd = curve.boundaries
            d_spline = UnivariateSpline(x, yb, s=0, ext=3)
            ax.plot(top, d_spline(top), 'o', color='red', label='peak tops')
            if len(bnd) > 0:
                b_spline = UnivariateSpline(x, base, s=0, ext=3)
                ax.plot(bnd, b_spline(bnd), 'o', color='green', label='valley bottoms')
            ax.legend()

        fig.tight_layout()
        fig.subplots_adjust(top=0.85)
    plt.show()

def save_mapped_curves(mapper, plot=False):
    import os
    import numpy as np
    from molass_legacy._MOLASS.SerialSettings import get_setting
    analysis_folder = get_setting('analysis_folder')

    x = mapper.x_curve.x
    mapped_uv_vector = mapper.make_uniformly_scaled_vector(scale=1)

    if plot:
        with plt.Dp():
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            for k, (name, curve) in enumerate([("uv_curve", mapper.a_curve), ("xr_curve", mapper.x_curve)]):
                ax = ax1 if k == 0 else ax2
                ax.plot(curve.x, curve.y, label=name)

            ax2.plot(x, mapped_uv_vector, label="mapped uv")
            ax2.legend()
            fig.tight_layout()
            plt.show()

    for name, curve in [("uv_curve", mapper.a_curve), ("xr_curve", mapper.x_curve)]:
        file = os.path.join(analysis_folder, name + ".dat")
        np.savetxt(file, np.array([curve.x, curve.y]).T)

    file = os.path.join(analysis_folder, "mapped_uv_curve.dat")
    np.savetxt(file, np.array([x, mapped_uv_vector]).T)



"""
    Theory.SphericalFit.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
from scipy.interpolate import UnivariateSpline
from scipy.optimize import minimize
from matplotlib.patches import Rectangle
import molass_legacy.KekLib.DebugPlot as plt
from PlotUtils import align_yaxis_np
from MatrixData import simple_plot_3d
from SvdDenoise import get_denoised_data
from .Rg import compute_corrected_Rg, compute_corrected_Rg_impl
from .SolidSphere import phi
from Rank.Synthesized import synthesized_data
from .SolidSphere import get_boundary_params, get_boundary_params_simple
from DataUtils import get_in_folder

CONC_PICK_RATIO = 0.8

def conc_pick(i):
    return int(i*CONC_PICK_RATIO)

def estimate_L(qv, M, Rg):
    R = np.sqrt(5/3)*Rg
    M_ = get_denoised_data(M, rank=2)
    Aq = phi(qv,R)**2
    Bq = -Aq*phi(qv,2*R)
    P_ = np.array([Aq, Bq]).T
    C_ = np.linalg.pinv(P_) @ M_

    c = C_[0,:]
    lc2 = C_[1,:]
    L_ = (lc2[np.newaxis,:]) @ np.linalg.pinv(c[np.newaxis,:]**2)

    return L_[0,0], C_

def debug_plot(c, qv, Aq, Bq, bq_):
    plt.push()
    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(14,7))
    ax1.plot(c)
    ax2.set_yscale('log')
    ax2.plot(qv, Aq)
    ax2t = ax2.twinx()
    ax2t.plot(qv, Bq)
    qv_ = qv[0:len(bq_)]
    ax2t.plot(qv_, bq_, color='pink')
    fig.tight_layout()
    plt.show()
    plt.pop()

def estimate_LR(arg_qv, arg_M, Rg, b1, debug=False):
    i = bisect_right(arg_qv, b1)
    qv = arg_qv[0:i]
    M = arg_M[0:i,:]

    R = np.sqrt(5/3)*Rg
    M_ = get_denoised_data(M, rank=2)
    Aq = phi(qv,R)**2

    c = arg_M[conc_pick(i),:]

    Pw = np.zeros((len(qv),2))
    Pw[:,0] = Aq

    def objective(x):
        rscale, bscale = x
        bq = -Aq*phi(qv, rscale*R)
        Cw = np.array([c, c**2])
        Pw = M_ @ np.linalg.pinv(Cw)
        return np.linalg.norm(bq - bscale*Pw[:,1])

    res = minimize(objective, (2, 1), bounds=((2, 10), (1e-3, None)))
    rscale, bscale = res.x
    score = res.fun/np.linalg.norm(-Aq*phi(qv, rscale*R))
    print('res.x=', res.x, 'res.fun=', res.fun, 'score=', score)

    Bq = -Aq*phi(qv, rscale*R)

    if debug:
        Cw = np.array([c, c**2])
        Pw = M_ @ np.linalg.pinv(Cw)
        bq_ = bscale*Pw[:,1]
        debug_plot(c, qv, Aq, Bq, bq_)

    P_ = np.array([Aq, Bq]).T
    C_ = np.linalg.pinv(P_) @ M_
    c = C_[0,:]
    lc2 = C_[1,:]
    L_ = (lc2[np.newaxis,:]) @ np.linalg.pinv(c[np.newaxis,:]**2)

    return L_[0,0], C_, rscale, bscale, score

def demo(parent, sd, pno=0, use_estimated_c=False, use_better_estimate=True, debug=False):
    M, E, qv, ecurve = sd.get_xr_data_separate_ly()

    range_ = ecurve.get_ranges_by_ratio(0.5)[pno]
    f_ = range_[0]
    p_ = range_[1]
    t_ = range_[2]

    if True:
        f, t = f_, t_
    else:
        if p_ < M.shape[1]*0.7:
            f, t = p_, t_
        else:
            f, t = f_, p_

    eslice = slice(f,t)

    Rg, gf, gt, guinier_truncated = compute_corrected_Rg(sd, ecurve, pno, qv, M[:,eslice], E[:,eslice])
    aslice = slice(0, gt)

    fig = plt.figure(figsize=(21,7))
    ax1 = fig.add_subplot(131, projection='3d')
    ax2 = fig.add_subplot(132)
    ax3 = fig.add_subplot(133)

    fig.suptitle("Spherical Fitting Demo for " + get_in_folder(), fontsize=30)
    ax1.set_title("Guinier Region" + guinier_truncated, fontsize=20)
    ax2.set_title("Estimated Concentration Curve", fontsize=20)
    ax3.set_title("Scattering Curves and Interparticle Effects", fontsize=20)

    simple_plot_3d(ax1, M, x=qv, alpha=0.3)
    if False:
        gslice = slice(gf, gt)
        simple_plot_3d(ax1, M[gslice,:], x=qv[gslice], color='green', edgecolor='green', alpha=0.5)

    y = sd.jvector

    for i in [gf, gt]:
        x = np.ones(len(y))*qv[i]
        z = M[i,:]
        ax1.plot(x, y, z, color='green')

    if use_better_estimate:
        b1, b2, k = get_boundary_params_simple(Rg)
        L, C_, rscale, bscale, score = estimate_LR(qv, M[:,eslice], Rg, b1, debug=debug)
        print('L=', L, 'bscale=', bscale)
        c = C_[0,:]
        kc2 = C_[1,:]
    else:
        L, C_ = estimate_L(qv[aslice], M[aslice,eslice], Rg)
        print('L=', L)
        c = C_[0,:]
        kc2 = C_[1,:]
        rscale = 2
        bscale = 1
        b1, b2, k = get_boundary_params(Rg)

    R = np.sqrt(5/3)*Rg
    aq_ = phi(qv,R)**2
    # rq_ = -L * phi(qv,rscale*R)
    rq_ = -phi(qv,rscale*R)
    bq_ = aq_ * rq_

    print('boundary=', b1)
    M_, _ = synthesized_data(qv, M[:,eslice], Rg, cd=2, boundary=b1, k=k)

    i = bisect_right(qv, b1)
    if use_estimated_c:
        c_ = c
    else:
        # c_ = M_[i,:]    # this makes a spike at the boundary
        c_ = M[conc_pick(i),eslice]

    print('max c =', np.max(c))

    qlim = len(qv)
    C = np.array([c_, c_**2])
    P = M_[0:qlim,:] @ np.linalg.pinv(C)
    aq = P[:,0]
    scale = aq[0]
    print('aq scale=', scale)

    aq = P[:,0]/scale
    if use_better_estimate:
        bq = P[:,1]*bscale
    else:
        bq = P[:,1]/(scale**2)      # not sure if this is correct

    if debug:
        debug_plot(c_, qv, aq_, bq_, bq)

    j = np.argmax(c)
    ey = ecurve.y*c[j]/ecurve.y[f+j]

    ax2.plot(y, ey, ':', color='orange')
    y_ = y[eslice]
    ax2.plot(y_, c, color='orange', label='$ c_{1,*} $')
    ax2t = ax2.twinx()
    ax2t.grid(False)
    ax2t.plot(y_, L*c**2, alpha=0.3, label='$ L(=%.3g) \cdot {c_{1,*}}^2 $' % L)
    ax2t.plot(y_, kc2, alpha=0.3, label='$ c_{2,*} $')
    align_yaxis_np(ax2, ax2t)

    leg = ax2.legend(loc='upper left', fontsize=16)
    fig.canvas.draw()
    bounds = leg.get_frame().get_bbox().bounds
    x, y = ax2.transAxes.inverted().transform(bounds[0:2])
    print('(x,y)=', (x,y))
    if x < 0.5:
        pos = (0,y)
        side = 'left'
    else:
        pos = (1,y)
        side = 'right'
    ax2t.legend(bbox_to_anchor=pos, loc='upper ' + side, fontsize=16)

    ymin, ymax = ax2.get_ylim()
    p = Rectangle(
            (f, ymin),      # (x,y)
            t - f,          # width
            ymax - ymin,    # height
            facecolor   = 'cyan',
            alpha       = 0.2,
        )
    ax2.add_patch(p)

    ax3.set_yscale('log')

    ax3.plot(qv, aq_, color='C0', label='sphere A(q)')
    ax3.plot(qv[0:qlim], aq, color='C1', label='sample A(q)')

    ymin, ymax = ax3.get_ylim()
    ax3.set_ylim(ymin, 1e2)

    ax3t = ax3.twinx()
    ax3t.grid(False)
    ax3t.plot(qv[0:qlim], bq, color='pink', label='sample B(q)')
    ax3t.plot(qv, bq_, color='cyan', label='shpere B(q)')
    ax3t.plot(qv, rq_, color='yellow', label='sphere B(q)/A(q)')

    ymin, ymax = ax3t.get_ylim()
    ymax_ = -0.2*ymin + 1.2*ymax
    ax3t.set_ylim(ymin, ymax_)

    ax3t.plot([b1, b1], [ymin, ymax_], ':', color='red', label='boundary')
    ax3t.plot([b2, b2], [ymin, ymax_], ':', color='gray', label='extinction')

    ax3.legend(loc='upper right', fontsize=16)
    ax3t.legend(loc='lower left', fontsize=16)

    fig.tight_layout()
    fig.subplots_adjust(top=0.85)
    plt.show()

def compute_bq_score(M, E, qv, ecurve, gslice, eslice, preRg):
    gf = gslice.start
    gt = gslice.stop
    aslice = slice(0, gt)
    Rg = compute_corrected_Rg_impl(qv, M[:,eslice], E[:,eslice], preRg, gf, gt)
    b1, b2, k = get_boundary_params_simple(Rg)
    L, C_, rscale, bscale, score = estimate_LR(qv, M[:,eslice], Rg, b1)
    return Rg, L, rscale, bscale, score



"""
    SerialDataUtils.py

    Copyright (c) 2018-2025, SAXS Team, KEK-PF
"""
import os
import re
import glob
import numpy as np
from molass_legacy.KekLib.BasicUtils import clear_dirs_with_retry
from molass_legacy.KekLib.NumpyUtils import np_loadtxt, np_savetxt
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting, set_path_length
from molass_legacy.SerialAnalyzer.DevSettings import get_dev_setting
from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker, log_exception

file_name_filter_re = re.compile( r'(UV|spectra)[\w-]*\.txt$', re.IGNORECASE )
file_name_re    = re.compile( r'\D(\d+).txt$' )
data_line_re    = re.compile( r'^\d' )

warned_folders = {}

def get_xray_files( in_folder, file_extension=None, name_check=True ):
    if in_folder is None:
        return []

    if file_extension is None:
        file_extension = get_setting( 'file_extension' )

    datafiles = sorted( glob.glob( in_folder + '/' + file_extension ) )
    if not name_check:
        return datafiles

    f_ext = file_extension.replace( '*', '\\' )
    xray_file_re    = re.compile( r'_(\d{5})\w*' + f_ext + '$' )

    ret_files = []
    err_files = []
    for k, f in enumerate( datafiles ):
        m = xray_file_re.search( f )
        if m:
            n = int(m.group(1))
            # assert n == k
            ret_files.append( f )
        else:
            err_files.append( f )

    if len(err_files) < len(ret_files):

        if len(err_files) > 0:
            already_warned = warned_folders.get(in_folder)
            suppress_warning = get_setting( 'suppress_warning' )
            test_pattern = get_setting("test_pattern")
            if already_warned is None and not suppress_warning and test_pattern is None:
                import molass_legacy.KekLib.OurMessageBox as MessageBox
                files = [ os.path.split( f )[1] for  f in err_files ]
                MessageBox.showwarning( "Warning", "irregular filenames " + str(files)
                                        + " in " + in_folder + " have been ignorged." )
                warned_folders[in_folder] = 1

        return ret_files
    else:
        return datafiles

def get_file_no( name ):
    m = file_name_re.search( name )
    if m:
        return int( m.group(1) )
    else:
        return None

def find_conc_files( in_folder ):
    files = filter( lambda x: file_name_filter_re.search( x ), glob.glob( in_folder + '/*.txt' ) )
    file_recs = map( lambda x: [ get_file_no( x ), x ], files )
    datafile_recs = sorted( file_recs )
    return datafile_recs

def get_uv_filename( uv_folder, glob=None ):
    disable_uv_data = get_setting('disable_uv_data')
    if disable_uv_data:
        return None

    datafile_recs  = list( find_conc_files( uv_folder ) )
    # print( 'datafile_recs=', datafile_recs )
    if len(datafile_recs) > 0:
        dir_, file = os.path.split( datafile_recs[0][1] )
        if len(datafile_recs) == 1:
            return file
        else:
            return '*.txt'
    else:
        return try_to_find_uv_in_any_txt(uv_folder)

def try_to_find_uv_in_any_txt(uv_folder):
    ret_file = None
    for file in glob.glob(uv_folder + '/*.txt'):
        try:
            found =False
            value_count = 0
            fh = open(file, 'r', encoding='cp932')
            for k, line in enumerate(fh):
                # print([k], line)
                m = data_line_re.match(line)
                if m:
                    vec = np.array([float(v) for v in line.split("\t")[:-1]])
                    if vec[0] > 190 and len(vec) > 100:
                        value_count += 1
                        if value_count > 10:
                            found = True
                            folder, ret_file = os.path.split(file)
                            break
            fh.close()
        except:
            log_exception(None, "try_to_find_uv_in_any_txt: ")
            continue
    return ret_file

def get_mtd_filename(in_folder):
    for folder in [in_folder, os.path.abspath(in_folder + r'\..')]:
        files = glob.glob(folder + r'\*.mtd')
        if len(files) > 0:
            break

    if len(files) == 0:
        name = None
    else:
        name = files[0]
    return name

def load_uv_file(file, column_header=False, return_dict=False):
    fh = open( file, 'r', encoding='cp932' )

    comment_lines = []

    def generator( fh ):
        for line in fh:
            if data_line_re.match( line ):
                yield line
            else:
                comment_lines.append( line )

    data = np.loadtxt( generator( fh ) )
    fh.close()

    # print( 'comment_lines=', comment_lines )

    uv_date_line_re = re.compile( r'^Date\s+(\d+)年(\d+)月(\d+)日' )
    uv_device_line_re = re.compile( r'^Spectrometers:\s+(\w+)' )

    measurement_date = None
    uv_device_no = None

    for line in comment_lines:
        m = uv_date_line_re.match(line)
        if m:
            measurement_date = int(m.group(1) + m.group(2) + m.group(3))
            # print('uv measurement_date=', measurement_date)
            continue
        m = uv_device_line_re.match(line)
        if m:
            uv_device_no = m.group(1)
            break

    set_setting("uv_device_no", uv_device_no)
    set_path_length(measurement_date, uv_device_no)

    if return_dict:
        return dict(data=data,
                    comment_lines=comment_lines,
                    measurement_date=measurement_date,
                    uv_device_no=uv_device_no)

    if column_header:
        if len( comment_lines ) > 1:
            header = comment_lines[-2]
            col_header = header.split( '\t' )[0:-1]     # changed from [1:-1]
        else:
            # as in 00_model_*
            col_header = None
        # print( 'len(col_header)=', len(col_header) )
        # print( 'data.shape=', data.shape )
        return data, col_header
    else:
        return data

def load_intensity_files( in_folder, logger=None, return_comments=False, debug=False ):
    datafiles = get_xray_files( in_folder )
    data_array, comments = load_xray_files( datafiles, debug=debug )
    if return_comments:
        return data_array, datafiles, comments
    else:
        return data_array, datafiles

def load_xray_files(datafiles_, return_lacking_info=False, qv=None, uniform_qv=True, logger=None, debug=False ):
    data_list = []
    comments = None
    encoding = 'cp932'
    size_list = []
    for k, file in enumerate(datafiles_):
        try:
            data, _ = serial_np_loadtxt( file, encoding=encoding )
            if debug:
                print([k], file, data.shape)

            size_list.append(data.shape[0])

            if comments is None:
                comments = _

        except UnicodeDecodeError as exc:
            etb = ExceptionTracebacker()
            encoding = 'utf-8'
            retry_message = "retrying with encoding = '%s'" % encoding
            if logger is None:
                print(etb)
                print(retry_message)
            else:
                logger.warning(retry_message)
            data, _ = serial_np_loadtxt( file, encoding=encoding )
        except Exception as exc:
            print( "Error: Can't read " + file + '\n\t' + str(exc))
            # TODO: exe log gui
            etb = ExceptionTracebacker()
            print(etb)
            continue
        data_list.append( data )

    if uniform_qv:
        size_array = np.array(size_list)
        minn = np.min(size_array)
        maxn = np.max(size_array)
        if minn < maxn or qv is not None:
            data_list, lacking_list, minn_qv = convert_to_the_least_shape(data_list, size_array, minn, maxn, qv=qv)
        else:
            lacking_list = []
            minn_qv = None
    else:
        pass

    measurement_date = None
    sangler_version = None
    if comments is not None:
        for line in comments:
            if line.find('Date') > 0:
                date_re = re.compile(r'(\d+/\d+/\d+)')
                m = date_re.search(line)
                if m:
                    measurement_date = int(m.group(1).replace('/', ''))
            elif line.find('SAngler Version') > 0:
                version_re = re.compile(r':\s(\S+)')
                m = version_re.search(line)
                if m:
                    sangler_version = m.group(1)
                    break

    set_setting('measurement_date', measurement_date)
    set_setting('sangler_version', sangler_version)
    if return_lacking_info:
        return np.array( data_list ), comments, (minn, maxn, lacking_list, minn_qv)
    else:
        return np.array( data_list ), comments

def convert_to_the_least_shape(data_list, size_array, minn, maxn, qv=None):
    import logging
    logger = logging.getLogger(__name__)
    logger.warning("Converting the input data set by adjusting to the intersecting q-values with (minn, maxn)=(%d, %d)", minn, maxn)

    wh = np.where(size_array == maxn)[0]
    maxn_qv = data_list[wh[0]][:,0]

    if qv is None:
        wh = np.where(size_array == minn)[0]
        minn_qv = data_list[wh[0]][:,0]
    else:
        minn_qv = qv

    while True:
        ok = True
        ret_list = []
        lacking_list = []

        for k, data in enumerate(data_list):
            isect_qv, _, isect_ind = np.intersect1d(minn_qv, data[:,0], return_indices=True)
            if data.shape[0] < maxn:
                # 
                d = np.setdiff1d(np.union1d(maxn_qv, data[:,0]), np.intersect1d(maxn_qv, data[:,0]))
                lacking_list.append((k, data.shape[0], list(d)))
            if np.array_equal(isect_qv, minn_qv):
                ret_list.append(np.array([minn_qv, data[isect_ind,1], data[isect_ind,2]]).T)
            else:
                logger.warning("retry at %d-th file.", k)
                ok = False
                minn_qv = isect_qv
                break

        if ok:
            break

    logger.warning("Files with lacking q-values were, in (file no, size, lacking Q's) list,")
    for pair in lacking_list:
        logger.warning("\t%s", str(pair))

    logger.warning("The number of q-values has been changed from %d to %d finally.", maxn, len(minn_qv))
    set_setting('found_lacking_q_values', 1)

    return ret_list, lacking_list, minn_qv

def load_uv_array( conc_folder, conc_file=None, column_header=False ):
    disable_uv_data = get_setting('disable_uv_data')

    if disable_uv_data:
        datafile_recs = []
        conc_file = None
    else:
        if conc_file is None:
            if not os.path.exists( conc_folder ):
                raise Exception( conc_folder + ' does not exist!' )

            datafile_recs = find_conc_files( conc_folder )
            # print( 'datafile_recs=', datafile_recs )
            conc_file = conc_folder + '/*.txt'
        else:
            if conc_folder is None:
                conc_file_path = conc_file
            else:
                conc_file_path = os.path.join( conc_folder, conc_file  )
            if not os.path.exists( conc_file_path ):
                raise Exception( conc_file_path + ' does not exist!' )
            datafile_recs = [ [ 1, conc_file_path ] ]
            conc_file = conc_file_path

    data_array = None
    col_header = None
    lvector = None

    for no, file in datafile_recs:
        if col_header is None:
            data, col_header = load_uv_file( file, column_header=True )
        else:
            data = load_uv_file( file )

        if data_array is None:
            lvector     = data[ :, 0 ] 
            data_array  = data[ :, 1: ] 
        else:
            data_array = np.hstack( ( data_array, data[ :, 1: ] ) )

    if column_header:
        return data_array, lvector, conc_file, col_header
    else:
        return data_array, lvector, conc_file

def sa_uv_loadtxt(path):
    folder, file = os.path.split(path)
    ret = load_uv_array(folder, file, column_header=True)
    return ret[0], ret[1], ret[3]

def sa_uv_savetxt(path, vector, data, header):
    with open(path, "w", newline="\n") as fh:
        fh.write("編集データ\n")
        fh.write(">>>>>>>>>>>>>> 連続計測（領域波長）Data Start<<<<<<<<<<<<\n")
    np_savetxt(path, np.vstack([vector, data.T]).T, mode="a", column_names=header)
    with open(path, "a", newline="\n") as fh:
        fh.write(">>>>>>>>>>>>>> Data End <<<<<<<<<<<<\n")

def serial_np_loadtxt( filename, encoding='cp932' ):
    if filename.find( '.int' ) > 0:
        return np_loadtxt_intensity( filename, encoding=encoding )
    else:
        return np_loadtxt( filename, encoding=encoding )

def np_loadtxt_intensity( filename, encoding='cp932' ):
    fh = open( filename, encoding=encoding )

    comment_lines = []

    def generator( fh ):
        for line in fh:
            if len( comment_lines ) < 3:
                comment_lines.append( line )
            else:
                yield line

    try:
        ret = np.loadtxt( generator( fh ) )
        array = ret[ :, 1:4 ]
    except:
        etb = ExceptionTracebacker()
        etb.log()
        array = None

    fh.close()
    return array, comment_lines

def save_xray_base_profiles( serial_data, corrected_base ):
    print( 'save_xray_base_profiles: corrected_base.shape=', corrected_base.shape )

    qvector = serial_data.qvector
    qsize   = len(qvector)
    qvector_ = qvector.reshape( (1, qsize) )
    x_array = serial_data.intensity_array

    xb_folder           = get_dev_setting( 'xb_folder' )
    base_file_postfix   = get_dev_setting( 'base_file_postfix' )
    postfix_ext         = base_file_postfix + '.dat'

    clear_dirs_with_retry( [xb_folder] )

    for i, file in enumerate( serial_data.datafiles ):
        # print( file )
        d, f = os.path.split( file )
        f_  = f.replace( '.dat', postfix_ext )
        out_file = os.path.join( xb_folder, f_ ).replace( '/', '\\' )
        base_data   = np.vstack( [ qvector_, corrected_base[:,i].reshape( (1, qsize) ), x_array[i,:,2] ] ).T
        np_savetxt( out_file, base_data )



"""
    SecTheory.TriporeSimulation.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import os
import sys
import numpy as np
from matplotlib.gridspec import GridSpec
from scipy.interpolate import UnivariateSpline

def demo():
    import molass_legacy.KekLib.DebugPlot as plt
    from SecTheory.SecCF import gec_monopore_phi, gec_dipore_phi, gec_tripore_phi
    from SecTheory.SecPDF import FftInvPdf
    from SecTheory.ColumnIllust import plot_column_illust, plot_column_legend

    monopore_pdf = FftInvPdf(gec_monopore_phi)
    dipore_pdf = FftInvPdf(gec_dipore_phi)
    tripore_pdf = FftInvPdf(gec_tripore_phi)

    x = np.arange(200)

    me = 0.1
    mp = 1.5
    rp0   =  76
    rp2_1 =  90
    rp2_2 =  40
    rp3_1 = 120
    rp3_2 =  76
    rp3_3 =  40

    rg = 35
    N = 171
    T = 0.63
    t0 = 20

    rg_array = np.array([70, 65, 60, 55, 50, 45, 40, 35, 30, 25, 20, 15, 10, 5, 3, 1])
    rho0 = rg_array/rp0
    rho0[rho0 > 1] = 1
    np0 = N*(1 - rho0)**me
    tp0 = T*(1 - rho0)**mp

    rho2_1 = rg_array/rp2_1
    rho2_1[rho2_1 > 1] = 1
    np2_1 = N*(1 - rho2_1)**me
    tp2_1 = T*(1 - rho2_1)**mp

    rho2_2 = rg_array/rp2_2
    rho2_2[rho2_2 > 1] = 1
    np2_2 = N*(1 - rho2_2)**me
    tp2_2 = T*(1 - rho2_2)**mp

    rho3_1 = rg_array/rp3_1
    rho3_1[rho3_1 > 1] = 1
    np3_1 = N*(1 - rho3_1)**me
    tp3_1 = T*(1 - rho3_1)**mp

    rho3_2 = rg_array/rp3_2
    rho3_2[rho3_2 > 1] = 1
    np3_2 = N*(1 - rho3_2)**me
    tp3_2 = T*(1 - rho3_2)**mp

    rho3_3 = rg_array/rp3_3
    rho3_3[rho3_3 > 1] = 1
    np3_3 = N*(1 - rho3_3)**me
    tp3_3 = T*(1 - rho3_3)**mp

    y1_list = []
    confpoints1 = []
    y2_list = []
    confpoints2 = []
    y3_list = []
    confpoints3 = []

    k = 0
    for np0_, tp0_, np1_, tp1_, np2_, tp2_, np3_1_, tp3_1_, np3_2_, tp3_2_, np3_3_, tp3_3_ in zip(np0, tp0, np2_1, tp2_1, np2_2, tp2_2, np3_1, tp3_1, np3_2, tp3_2, np3_3, tp3_3):
        y1 = monopore_pdf(x,  np0_, tp0_, t0)
        y1_list.append(y1)
        j = np.argmax(y1)
        confpoints1.append((x[j], rg_array[k]))

        y2 = dipore_pdf(x, 0.5*np1_, tp1_, 0.5*np2_, tp2_, t0)
        y2_list.append(y2)
        j = np.argmax(y2)
        confpoints2.append((x[j], rg_array[k]))

        y3 = tripore_pdf(x, 1/3*np3_1_, tp3_1_, 1/3*np3_2_, tp3_2_, 1/3*np3_3_, tp3_3_, t0)
        y3_list.append(y3)
        j = np.argmax(y3)
        confpoints3.append((x[j], rg_array[k]))

        k += 1

    confpoints1 = np.array(confpoints1)
    confpoints2 = np.array(confpoints2)
    confpoints3 = np.array(confpoints3)

    def create_spline(confpoints):
        return UnivariateSpline(confpoints[:,0], confpoints[:,1], s=0)

    spline1 = create_spline(confpoints1)
    spline2 = create_spline(confpoints2)
    spline3 = create_spline(confpoints3)

    x_ = x[t0+8:130]

    with plt.Dp():
        fig, axes = plt.subplots(nrows=2, ncols=4, figsize=(18, 9))
        ax1, ax2, ax3, ax4 = axes[0,:]

        fig.suptitle("Comparison of Conformance Curves among Monopore, Dipore and Tripore Columns", fontsize=20)
        ax1.set_title("Monopore Column with Poresize=%g" % rp0, fontsize=16)
        ax2.set_title("Dipore Column with Poresizes=%g, %g" % (rp2_1, rp2_2), fontsize=16)
        ax3.set_title("Tripore Column with Poresizes=%g, %g, %g" % (rp3_1, rp3_2, rp3_3), fontsize=16)
        ax4.set_title("Overlay of the Conformance Curves", fontsize=16)

        ylimits = []
        for ax, y_list, spline, confpoints, color1, color2, name in [
                    (ax1, y1_list, spline1, confpoints1, "yellow", "red", "Monopore"),
                    (ax2, y2_list, spline2, confpoints2, "cyan", "green", "Dipore"),
                    (ax3, y3_list, spline3, confpoints3, "orange", "blue", "Tripore")]:

            axt = ax.twinx()
            axt.grid(False)

            for y, rg_ in zip(y_list, rg_array):
                ax.plot(x, y, alpha=0.5, label="$R_g=%.3g$" % rg_)

            for k, ax_ in enumerate([axt, ax4]):
                label = None if k == 0 else name
                ax_.plot(x_, spline(x_), color=color1, lw=3, alpha=0.5, label=name)
                ax_.plot(*confpoints.T, "o", color=color2, alpha=0.5)

            # ax.legend()
            ylimits.append(ax.get_ylim())

        ax4.legend()
        ax4.set_xlim(0, 200)

        ylimits = np.array(ylimits)
        ymin = np.min(ylimits[:,0])
        ymax = np.max(ylimits[:,1])
        for ax in ax1, ax2, ax3:
            ax.set_ylim(ymin, ymax)

        plot_column_illust(axes[1,0], [rp0], ["white"])
        plot_column_illust(axes[1,1], [rp2_1, rp2_2], ["orange", "yellow"])
        plot_column_illust(axes[1,2], [rp3_1, rp3_2, rp3_3], ["pink", "white", "yellow"])
        plot_column_legend(axes[1,3], [rp3_1, rp2_1, rp3_2, rp3_3], ["pink", "orange", "white", "yellow"])

        fig.tight_layout()
        plt.show()

if __name__ == '__main__':
    this_dir = os.path.dirname(os.path.abspath(__file__))
    lib_dir = os.path.dirname(this_dir)
    sys.path.append(lib_dir)

    import seaborn
    seaborn.set()
    import molass_legacy.KekLib

    demo()



# coding: utf-8
"""
    KPeaks.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np

def kmeans(k, X, max_iter=300):
    X_size,n_features = X.shape
    centroids  = X[np.random.choice(X_size,k)]
    new_centroids = np.zeros((k, n_features))
    cluster = np.zeros(X_size)

    for epoch in range(max_iter):
        for i in range(X_size):
            distances = np.sum((centroids - X[i]) ** 2, axis=1)
            cluster[i] = np.argsort(distances)[0]
        for j in range(k):
            new_centroids[j] = X[cluster==j].mean(axis=0)
        if np.sum(new_centroids == centroids) == k:
            print("break")
            break
        centroids =  new_centroids
    return cluster

class KPeaks:
    def __init__(self, k, X, max_iter=300):
        nrows, ncols = X.shape
        centroids  = X[np.random.choice(X_size,k)]



# coding: utf-8
"""
    GuinierRegion.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from bisect import bisect_right
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from MatrixData import simple_plot_3d
from DataUtils import get_in_folder

MARGIN_WIDTH = 0.13     # angstrom⁻¹

class GuinierRegionInspector(Dialog):
    def __init__(self, parent, dialog, data, i_smp, eslice):
        self.logger = logging.getLogger(__name__)
        self.dialog = dialog
        sd = dialog.serial_data
        pre_rg = dialog.pre_rg
        self.sd = sd
        self.sg = pre_rg.sg
        qv = sd.qvector
        margin  = int(round(MARGIN_WIDTH/(qv[1] - qv[0])))
        self.i_smp = i_smp
        self.i_stop = max(i_smp, self.sg.guinier_stop) + margin
        self.eslice = eslice
        self.aslice = aslice = slice(0, self.i_stop)
        self.gqv = qv[aslice]
        self.j0 = eslice.start
        self.gdata = data[aslice,eslice].copy()
        canvas2, canvas3 = self.dialog.frames[0].canvases[1:3]
        self.ptx = canvas2.opos
        self.ecurve = canvas2.ecurve
        self.asx = canvas3.x[canvas3.restrict_info.start]
        self.x = canvas3.x[aslice]      # same as self.gqv
        self.xx = self.x**2
        self.y = canvas3.y[aslice]
        self.compute_corrected_info()

        Dialog.__init__(self, parent, "Guinier Region Inspector", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()

        fig = plt.figure(figsize=(18,7))
        self.fig = fig

        in_folder = get_in_folder()
        fig.suptitle("Guinier Region Inspection on %s" % in_folder, fontsize=20)
        ax1 = fig.add_subplot(131, projection="3d")
        ax2 = fig.add_subplot(132)
        ax3 = fig.add_subplot(133)

        self.draw_3d(ax1)
        self.determin_2d_fig_limit()
        self.draw_default_start(ax2)
        self.draw_corrected_start(ax3)

        fig.tight_layout()
        fig.subplots_adjust(top=0.85, bottom=0.1)

        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)

    def draw_3d(self, ax):
        ax.set_title("Guinier Region 3D View", fontsize=16)

        framenums = self.j0 + np.arange(self.gdata.shape[1])
        simple_plot_3d(ax, self.gdata, x=self.gqv, y=framenums)

        length = self.gdata.shape[1]
        y = np.arange(length) + self.j0
        i_smp = self.i_smp
        for i, label in [(self.sg.guinier_start, "Guinier Start"), (self.sg.guinier_stop, "Guinier Stop"), (i_smp, "SMP")]:
            x = np.ones(length)*self.gqv[i]
            z = self.gdata[i,:]
            ax.plot(x, y, z, color="orange" if i == i_smp else "cyan", label=label)

        x = self.gqv
        y = np.ones(len(x)) * self.ptx
        z = self.y
        ax.plot(x, y, z, color="green", label="Peak Top Ridge")

        ax.legend()

    def determin_2d_fig_limit(self):
        qmax = 2.0/self.sg.Rg
        self.q2max = qmax**2
        print("q2max=", self.q2max)
        i = bisect_right(self.gqv, qmax)
        self.ymin = np.log(self.y[i])

    def draw_default_start(self, ax):
        self.draw_start_impl(ax, "Default Range Start")

    def draw_corrected_start(self, ax):
        self.draw_start_impl(ax, "Corrected Range Start", info=self.corrected_info)

    def draw_start_impl(self, ax, title, info=None):

        if info is None:
            x = self.x
            y = self.y
            sg = self.sg
            asx = self.asx
            y_orig = None
        else:
            x, y, sg, asx = info
            y_orig = self.y

        ax.set_title(title, fontsize=16)

        ax.set_xlabel("$Q^2$")
        ax.set_ylabel("$Ln(I)$")

        xx = x**2
        if y_orig is not None:
            ax.plot(xx, np.log(y_orig), ":", color="gray")
        ax.plot(xx, np.log(y), color="green")

        _, ymax = ax.get_ylim()
        ymin = self.ymin
        ax.set_ylim(ymin, ymax)

        k = 0
        gx = x[[sg.guinier_start, sg.guinier_stop]]
        for x_, label in [(asx, "Range Start"), (gx[0], "Guinier Start"), (gx[1], "Guinier Stop")]:
            color = "yellow" if k == 0 else "cyan"
            x2 = x_**2
            ax.plot([x2, x2], [ymin, ymax], color=color, label=label)
            k += 1

        self.plot_guinier_line(ax, sg)

        xmin, xmax = ax.get_xlim()
        w = 0.3
        tx = xmin*(1-w) + xmax*w
        w = 0.6
        ty = ymin*(1-w) + ymax*w
        ax.text(tx, ty, "Range Start = $%.3g^2$" % asx, fontsize=20, alpha=0.5)

        ax.legend()

    def compute_corrected_info(self):
        from LPM import LPM_3d
        from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve
        from molass_legacy._MOLASS.SerialSettings import get_setting
        from .PreliminaryRg import PreliminaryRg
        from .GuinierLimit import GuinierLimit

        lpm = LPM_3d(self.gdata)
        corrected_data = lpm.data
        j = self.ptx - self.j0
        y = corrected_data[:,j]
        ey = np.average(corrected_data[self.i_smp-5:self.i_smp+6,:], axis=0)
        error_data = self.sd.intensity_array[self.eslice, self.aslice, 2].T
        e_curve = ElutionCurve(ey)
        pre_rg = PreliminaryRg(corrected_data, error_data, e_curve, self.gqv, self.i_stop)
        gl = GuinierLimit(corrected_data, e_curve, pre_rg, self.i_stop)
        rg_consistency = get_setting('acceptable_rg_consist')
        try:
            angle_start = gl.get_limit(rg_consistency)
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception
            log_exception(self.logger, "GuinierLimit.get_limit failed.", n=10)
            angle_start = 0

        asx = self.gqv[angle_start]

        if False:
            import molass_legacy.KekLib.DebugPlot as dplt
            from molass_legacy.Elution.CurveUtils import simple_plot
            dplt.push()
            fig, ax = dplt.subplots()
            ax.set_title("compute_corrected_info debug")
            simple_plot(ax, e_curve)
            dplt.show()
            dplt.pop()

        self.corrected_info = (self.x, y, pre_rg.sg, asx)

    def plot_guinier_line(self, ax, sg):
        # see ProofPlot.plot_guineier_inverval
        Rg, I0, f, t = sg.Rg, sg.Iz, sg.guinier_start, sg.guinier_stop

        x = self.xx
        b_ = np.log(I0)
        a_ = -(Rg**2 / 3)
        y0 = b_ + a_ * x[f]
        y1 = b_ + a_ * x[t]
        ax.plot( [ x[f], x[t] ], [ y0, y1 ], marker='o', color="red", label="Guinier Line")

        xmin, xmax = ax.get_xlim()
        ax.set_xlim(xmin/4 if xmin < 0 else xmin, min(xmax, self.q2max))

    def save_the_figure(self, file):
        self.fig.savefig(file)



"""
    CompositeInfo.py

    Copyright (c) 2023-2024, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.Optimizer.OptLrfInfo import get_ratio_cy_list

def convert_for_display(composites):
    ret_list = []
    for comp in composites:
        ret_list.append([i+1 for i in comp])
    return ret_list

class CompositeInfo:
    def __init__(self, num_components=None, composites=None):
        # note that CompositeInfo does not incluse the baseline

        if composites is None:
            if num_components is None:
                pass
            else:
                assert type(num_components) is int
                composites = [[j] for j in range(num_components)]

        self.composites = composites    # note that this includes baseline
        if composites is None:
            # will be loaded later
            pass
        else:
            self.check_num_elements()
        self.separate_eoii = get_setting("separate_eoii")
        self.separate_eoii_type = get_setting("separate_eoii_type")
        self.separate_eoii_flags = get_setting("separate_eoii_flags")

    def check_num_elements(self):
        k = 0
        valid_indeces = []
        really_composite = False
        for comp in self.composites:
            if len(comp) > 1:
                really_composite = True
            for j, i in enumerate(comp):
                assert i == k
                if j == 0:
                    valid_indeces.append(k)
                k += 1
        self.num_elements = k 
        self.valid_indeces = np.array(valid_indeces[:-1])   # excluding baseline
        self.really_composite = really_composite

    def get_num_components(self):
        return len(self.composites)         # including the baseline

    def get_num_substantial_components(self):
        return len(self.composites) - 1     # excluding the baseline

    def get_valid_rgs(self, rgs):
        return rgs[self.valid_indeces]

    def get_composite_penalty(self):
        if self.really_composite:
            pass
        else:
            return 0

    def __str__(self):
        return str(self.composites)

    def get_extra_num_components(self):
        return 1 if self.separate_eoii else 0

    def save(self, path):
        with open(path, "w") as fh:
            fh.write(str(self.composites))

    def load(self, path):
        from molass_legacy.KekLib.EvalUtils import file_eval
        self.composites = file_eval(path)
        self.check_num_elements()

    def compute_C_matrix(self, y, cy_list, eoii=False, ratio_interpretation=False):
        # note that cy_list are non-composite curves

        work_list = []
        for comp in self.composites:
            for k, j in enumerate(comp):
                if k == 0:
                    cy = cy_list[j].copy()
                else:
                    cy += cy_list[j]
            work_list.append(cy)

        if ratio_interpretation:
            assert self.separate_eoii_type == 0
            cy_list = get_ratio_cy_list(y, cy_list)
            ty = y

        if eoii:
            ty = np.sum(cy_list[:-1], axis=0)   # excluding baseline
            if self.separate_eoii_type == 1:                
                work_list.append(ty**2)
            elif self.separate_eoii_type == 2:
                # not yet supported
                for k, flag in enumerate(self.separate_eoii_flags):
                    if flag:
                        work_list.append(cy_list[k]**2)

        C = np.array(work_list)
        return C

    def get_composite_ids(self):
        ids = []
        for k, comp in enumerate(self.composites):
            for i in comp:
                ids.append(k)
        return ids

    def get_composite_show_flags(self):
        flags = []
        for k, comp in enumerate(self.composites):
            for j, i in enumerate(comp):
                if j == 0:
                    flags.append(True)
                else:
                    flags.append(False)
        return flags

    def get_composite_proportions(self, proportions):
        ret_proportions = []
        for k, comp in enumerate(self.composites[:-1]):     # excluding baseline
            p = 0
            for j, i in enumerate(comp):
                if i < self.num_elements:
                    p += proportions[i]
            ret_proportions.append(p)
        return ret_proportions

    def get_elutions(self, cy_list, pv=None):
        if pv is None:
            pv = np.ones(len(cy_list))
        ret_list = []
        for comp in self.composites:
            for k, j in enumerate(comp):
                if k == 0:
                    cy = pv[j]*cy_list[j]
                else:
                    cy += pv[j]*cy_list[j]
            ret_list.append(cy)
        if len(ret_list) < len(cy_list):
            # temporary support for eoii components
            ret_list += cy_list[len(ret_list):]
        return ret_list

    def get_elements_rgs(self, rg_params):
        # this is deprecated replaced by get_valid_rgs

        ret_rgs = rg_params.copy()
        k = 0
        for comp in self.composites:
            n = len(comp)
            if n > 1:
                ret_rgs[k:k+n] = np.mean(rg_params[k:k+n])
            k += n
        return ret_rgs

    def estimate_peak_points(self, x, elutions):
        # use get_elutions method for elutions
        # do not use this method in optimization because it is slow

        n = self.get_num_substantial_components()
        peak_points = []
        for cy in elutions[:n]:
            j = np.argmax(cy)
            peak_points.append((x[j], cy[j]))
        peak_points = np.array(peak_points)
        return peak_points



# coding: utf-8
"""
    SimpleSolver.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import os
import numpy as np
from scipy import optimize, stats
from lmfit import Parameters, minimize
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.KekLib.SciPyCookbook import smooth
from SimpleUnfolding import SimpleUnfolding

MAX_ITERATION_FMIN_CG = 10000
RT = 1

class SimpleSolver:
    def __init__(self):
        pass

    def solve(self, data, index=None):
        print(data.shape)
        self.data = data
        self.guess_init_C(index)

    def guess_init_C(self, index):
        self.x = x = np.arange(self.data.shape[1])
        self.y = y = self.data[index,:]
        sy = smooth(y)
        self.model = SimpleUnfolding()
        res = self.model.fit(x, sy, G_init=10)     # 0.1, 1, 10
        self.params = res.params

    def plot_guess(self, ax, title=None):
        # print('params=', [(k, p.value) for k, p in self.params.items()])

        x = self.x
        y = self.y
        pf = self.model.compute_Pf(self.params, x)
        yf = self.model.compute_yf(self.params, pf, x)
        yu = self.model.compute_yu(self.params, pf, x)

        axt = ax.twinx()
        axt.grid(False)

        if title is not None:
            ax.set_title(title, fontsize=16)
        ax.set_xlabel('Eno')
        ax.set_ylabel('Proportion')
        axt.set_ylabel('Intensity')

        ax.plot(x, pf, label='$P_F$')
        axt.plot(x, y, color='orange', label='given')
        axt.plot(x, yf, ':', label='$y_F$')
        axt.plot(x, yu, ':', label='$y_U$')
        axt.plot(x, yf+yu, ':', color='orange', label='$y_F+y_U$')

        ax.legend(bbox_to_anchor=(0, 0.7), loc='center left', fontsize=16)
        axt.legend(loc='upper right', fontsize=16)

    def guess_blank_location(self, slice_):
        G = self.params['G']
        m = self.params['m']
        cx = G/m
        xsize = slice_.stop - slice_.start
        pos_ratio = cx/xsize
        slope, intercept, r_value, p_value, std_err = stats.linregress(self.x, self.y)

        # print('pos_ratio=', pos_ratio, 'slope=', slope)

        if pos_ratio < 0.4  or (pos_ratio < 0.6 and slope > 0):
            side = 'right'
        else:
            side = 'left'
        return side

    def get_crossing_point(self):
        G = self.params['G']
        m = self.params['m']
        cp = G/m
        # print('G=', G, 'm=', m, 'cp=', cp)
        return int(cp)

    def get_fit_model(self):
        return SimpleUnfolding(params=self.params)



# coding: utf-8
"""

    ファイル名：    AppVersion.py

    処理内容：      アプリケーションのバージョン情報


"""
import platform

def synthesizer_version_string():
    return 'Synthesizer 0.4.6 (2025-03-12 python %s %s)' % ( platform.python_version(), platform.architecture()[0] )



# coding: utf-8
"""
    SecTools.CorMap.SchillingDist.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np

def schilling_prob(a, b, n, p):
    r = -np.log(n*(1-p))/np.log(p)
    cdf = lambda x: np.exp(- p**x )
    return cdf(b + 1 - r) - cdf(a - r)

def schilling_pdf(n, p, x):
    y = np.zeros(len(x))
    dx = (x[1] - x[0])/2        # assuming that x is regularly divided
    for k, x_ in enumerate(x):
        y[k] = schilling_prob(x_-dx, x_+dx, n, p)
        px = x_
    return y



"""
    Trimming.GuinierLimit.py

    Copyright (c) 2018-2025, SAXS Team, KEK-PF
"""
import numpy as np
import logging
from scipy import stats
from molass_legacy.SerialAnalyzer.AveragingDiscussion import extrapolate
from molass_legacy.GuinierAnalyzer.SimpleGuinierScore import compute_rg
from molass_legacy._MOLASS.SerialSettings import get_setting

RG_EST_ELUTION_WIDTH    = 50
RG_EST_ANGLE_WIDTH      = 20

class GuinierLimit:
    def __init__(self, matrix_data, e_curve, pre_rg, qlimit):
        self.logger = logging.getLogger( __name__ )
        self.matrix_data = matrix_data
        self.e_curve = e_curve
        self.pre_rg = pre_rg
        self.qlimit = qlimit
        self.limit = None

    def determine_safe_start(self, debug=False):
        sg = self.pre_rg.sg
        x = sg.x
        y = sg.y
        q_limit = self.qlimit
        stop = len(y) if q_limit is None else q_limit
        n = np.argmax(y[0:stop])
        q_limit = min(n, stop)
        if q_limit > 0:
            m = np.argmin(y[0:q_limit])
        else:
            m = 0
        if m > 1:
            regress_values = []
            for slice_ in [slice(0,m), slice(m,None)]:
                slope, _, r_value = stats.linregress( x[slice_], y[slice_] )[0:3]
                regress_values.append((slope, r_value))
            regress_values = np.array(regress_values)
            min_r_value = np.min(np.abs(regress_values[:,1]))
            if min_r_value > 0.6:
                safe_start = m
                self.logger.info("safe_start is determined to be at Q[%d]=%.3g", m, x[m])
            else:
                safe_start = 0
        else:
            safe_start = 0

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
            plt.push()
            fig, axes = plt.subplots(ncols=3, figsize=(21,7))
            fig.suptitle('determine_safe_start debug for ' + get_in_folder(), fontsize=30)
            for ax, slice_ in zip([axes[0], axes[2]], [slice(None, None), slice(0, self.qlimit)]):
                ax.set_yscale('log')
                ax.plot(x[slice_], y[slice_])
            ax2 = axes[2]
            ymin, ymax = ax2.get_ylim()
            ax2.set_ylim(ymin, ymax)
            x_ = x[safe_start]
            ax2.plot([x_, x_], [ymin, ymax], ':', color='yellow', linewidth=3)
            if m > 1:
                print('regress_values=', regress_values)
                print('min_r_value=', min_r_value)

            fig.tight_layout()
            fig.subplots_adjust(top=0.88)
            plt.show()
            plt.pop()

        return safe_start

    def compute_extrapolated_curve(self, safe_start, debug=False):
        default_eno = self.pre_rg.default_eno
        ewidth  = RG_EST_ELUTION_WIDTH//2
        estart  = max(0, default_eno - ewidth)
        estop   = min(self.matrix_data.shape[1], default_eno + ewidth)
        eslice  = slice(estart, estop)

        sg = self.pre_rg.sg
        guinier_start = sg.guinier_start
        aslice  = slice(safe_start, guinier_start + RG_EST_ANGLE_WIDTH*2)

        M   = self.matrix_data[aslice, eslice]

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            from molass_legacy.DataStructure.MatrixData import simple_plot_3d
            from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
            plt.push()
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
            ax.set_title(get_in_folder())
            simple_plot_3d(ax, M)
            plt.show()
            plt.pop()

        c   = self.e_curve.y[eslice]
        P   = extrapolate(M, c)
        y   = P[:,0]
        return y

    def get_safe_check_stop(self, y):
        width = 5
        stop = None
        for i in range(10, -1, -1):
            i_ = i + 1
            y_ = y[i_:i_+width]
            height = np.max(y_) - np.min(y_)
            ay = np.average(y_)
            ratio = abs(y[i] - ay)/height
            # print([i], "ratio=", ratio)
            if ratio > 2:
                # as with y[0] in 20230303/HasA 
                stop = i
                break
        if stop is None:
            stop = -1
        return stop

    def get_limit(self, acceptable_consistency, debug=False, fig_file=None, debug_auto=False):
        assert acceptable_consistency >= 0 and acceptable_consistency <= 1

        if acceptable_consistency == 0:
            return 0

        safe_start = self.determine_safe_start(debug=debug)
        zx_y = self.compute_extrapolated_curve(safe_start, debug=debug)
        log_zx_y = np.log(zx_y)

        sg = self.pre_rg.sg
        guinier_start = sg.guinier_start

        log_y = sg.log_y
        x2 = sg.x2

        acceptable_ratio = -np.log(acceptable_consistency)
        """
            i.e.,
            np.exp(-acceptable_ratio) == acceptable_consistency
        """

        def compute_extrapolated_rg(i):
            slice_ = slice(i, i+RG_EST_ANGLE_WIDTH)
            x_ = x2[slice_]
            y_ = log_zx_y[slice_]
            # slope, intercept, r_value, p_value, std_err = stats.linregress( x_, y_ )
            slope = stats.linregress( x_, y_ )[0]
            rg = compute_rg(slope)            
            return rg

        limit = i = guinier_start
        reliable_rg = compute_extrapolated_rg(i)
        check_stop = self.get_safe_check_stop(log_zx_y)

        for i in range(guinier_start+1, check_stop, -1):
            rg = compute_extrapolated_rg(i)
            # print([i], rg, diff_ratio)
            if rg > 0:
                diff_ratio = abs(rg - reliable_rg)/reliable_rg
                if diff_ratio <= acceptable_ratio:
                    limit = i
                else:
                    continue
            else:
                continue

        if debug:
            import time
            print("limit=", limit)
            i = guinier_start

            with plt.Dp():
                fig, axes = plt.subplots(ncols=3, figsize=(21, 7))
                ax0, ax1, ax2 = axes

                in_folder = get_setting('in_folder')
                consistency_text = ' with acceptable Rg-consistancy %g' % acceptable_consistency
                fig.suptitle(in_folder + consistency_text, fontsize=20)

                ax0.set_title("Log plot of the whole region", fontsize=16)
                ax0.set_yscale('log')
                ax0.plot(sg.x, sg.y)
                ax0.plot(sg.x[i], sg.y[i], 'o', color='red')

                ax1.set_title("Guinier plot of the whole region", fontsize=16)
                ax1.plot(x2, log_y)
                aslice = slice(safe_start, safe_start+len(zx_y))
                ax1.plot(x2[aslice], log_zx_y)

                xmin1, xmax1 = ax1.get_xlim()
                ymin1, ymax1 = ax1.get_ylim()

                if True:
                    dx = (xmax1 - xmin1)*0.0005
                    dy = (ymax1 - ymin1)*0.005
                    gx = x2[i]
                    gy = log_y[i]
                    # lower = max(0, i - RG_EST_ANGLE_WIDTH)
                    upper = min(len(x2)-RG_EST_ANGLE_WIDTH*2-1, i + RG_EST_ANGLE_WIDTH*2)
                    xmin2 = x2[0] - dx
                    xmax2 = x2[upper] + dx
                    ymin2 = min(log_y[limit], log_y[upper]) - dy
                    ymax2 = max(gy, np.max(log_zx_y)) + dy
                else:
                    xmin2 = xmin1*0.96 + xmax1*0.04
                    xmax2 = xmin1*0.92 + xmax1*0.08
                    ymin2 = ymin1*0.15 + ymax1*0.85
                    ymax2 = ymin1*0.02 + ymax1*0.98

                ax2.set_title("Guinier plot of the small angle region", fontsize=16)
                ax2.set_xlim(xmin2, xmax2)
                ax2.set_ylim(ymin2, ymax2)

                ax2.plot(x2, log_y, label='measured curve')
                ax2.plot(x2[aslice], log_zx_y, label='extrapolated curve')

                for k, ax in enumerate(axes[1:]):
                    ax.plot(x2[i], log_y[i], 'o', color='red', label='guinier start')
                    if k > 0:
                        x_ = x2[i]
                        ax.plot([x_, x_], [ymin2, ymax2], ':', color='red', linewidth=1)
                        i = limit
                        x_ = x2[i]
                        ax.plot(x_, log_y[i], 'o', color='yellow', label='possible start')
                        ax.plot(x_, log_zx_y[i], 'o', color='yellow')
                        ax.plot([x_, x_], [ymin2, ymax2], ':', color='yellow', linewidth=3)

                ax2.legend(fontsize=16)

                fig.tight_layout()
                fig.subplots_adjust( top=0.88 )
                plt.show(block=not debug_auto)
                if fig_file is not None:
                    fig.savefig( fig_file )

            if debug_auto:
                time.sleep(0.5)

        # limit =  int( guinier_start * probability )     # this is a dummy setting
        return limit



"""
    Simulative.BridgeUtils.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import logging
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Decomposer.FitRecord import FitRecord
from molass_legacy.Decomposer.ModelEvaluator import ModelEvaluator
from Selective.PeakProxy import PeakProxy
from molass_legacy.QuickAnalysis.ModeledPeaks import get_modeled_peaks_impl
from molass_legacy.Decomposer.UnifiedDecompResult import UnifiedDecompResult

IGNORABLE_PROPORTION = 0.01

def make_traditional_fitrecs(x, y, model, peaks, debug=False):
    opt_recs = []
    kno = 0
    area = np.sum(y)
    for k, params in enumerate(peaks):
        cy = model(x, params)
        j = np.argmax(cy)
        peak = PeakProxy(top_x=x[j], top_y=cy[j])
        carea = np.sum(cy)
        area_prop = carea/area
        if area_prop > IGNORABLE_PROPORTION:
            fit_rec = FitRecord(kno, ModelEvaluator(model, params, sign=1), 0, peak, area=carea)
            opt_recs.append(fit_rec)
            kno += 1
        else:
            logger = logging.getLogger(__name__)
            logger.info("element with area_prop=%g <= %g has been ingored in decompose_curve.", area_prop, IGNORABLE_PROPORTION)

    return opt_recs

def decompose_by_bridge(x, y, uv_y, model, traditional_info, debug=False):
    from Selective.BridgeUtils import make_traditional_fitrecs
    num_peaks = traditional_info.num_peaks
    mapper = traditional_info.mapper
    print("num_peaks=", num_peaks)

    if debug:
        def plot_curves(ax, x, y, recs=None):   
            ax.plot(x, y, label="data")
            if recs is not None:
                cy_list = []
                for k, rec in enumerate(recs):
                    cy = rec.evaluator(x)
                    ax.plot(x, cy, ":", label="component-%d" % k)
                    cy_list.append(cy)
                ty = np.sum(cy_list, axis=0)
                ax.plot(x, ty, ":", color="red", label="model total")
            ax.legend()

        with plt.Dp():
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            fig.suptitle("decompose_by_bridge input")
            plot_curves(ax1, x, uv_y)
            plot_curves(ax2, x, y)
            fig.tight_layout()
            plt.show()

    # task: make get_modeled_peaks_impl support other models than EGHA
    uv_peaks, xr_peaks = get_modeled_peaks_impl(1, 0, x, uv_y, x, y, num_peaks, exact_num_peaks=num_peaks, affine=True)
    opt_recs = make_traditional_fitrecs(x, y, model, xr_peaks)
    opt_recs_uv = make_traditional_fitrecs(x, uv_y, model, uv_peaks)

    if debug:
        with plt.Dp():
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            fig.suptitle("decompose_by_bridge results")
            plot_curves(ax1, x, uv_y, recs=opt_recs_uv)
            plot_curves(ax2, x, y, recs=opt_recs)
            fig.tight_layout()
            plt.show()

    result = UnifiedDecompResult(
                xray_to_uv=None,
                x_curve=mapper.x_curve, x=x, y=y,
                opt_recs=opt_recs,
                max_y_xray = np.max(y),     # get this from traditional_info
                model_name=model.get_name(),
                decomposer=None,            # what is this used for?
                uv_y=uv_y,
                opt_recs_uv=opt_recs_uv,
                max_y_uv = np.max(uv_y),    # get this from traditional_info
                )

    result.set_area_proportions()
    result.set_proportions()    # do not remove_unwanted_elements, set proportions only.

    return result


"""

    Rank.SrrStats.py

    Copyright (c) 2023, SAXS Team, KEK-PF

"""
import numpy as np
import matplotlib.pyplot as plt
import seaborn
seaborn.set()

def demo(srr_csv):
    srr_list = []
    with open(srr_csv) as fh:
        for line in fh:
            str_values = line[:-1].split(',')
            values = [np.nan if s == "None" else float(s) for s in str_values[-6:]]
            srr_list.append(values)

    srr_array = np.array(srr_list)

    if srr_array.shape[1] > 2:
        fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(18,5))
    else:
        fig, ax2 = plt.subplots()

    ax1.set_title("SCD Distribution", fontsize=16)
    scd = srr_array[:,0]
    ax1.hist(scd, bins=40, label="SCD")
    ymin, ymax = ax1.get_ylim()
    ax1.set_ylim(ymin, ymax)
    for name, scd_ in [("20180526/GI", 5.37495), ("20180526/OA", 10.5122), ("20190527/EndoAB", 0.469393)]:
        ax1.plot([scd_, scd_], [ymin, ymax], ":", label="%s: SCD=%.3g" % (name, scd_))
    ax1.legend()
    ax1.set_xlabel("SCD")
    ax1.set_ylabel("Counts")

    ax2.set_title("Two Types of SRR Distributions", fontsize=16)

    ax2.hist(srr_array[:,2], bins=40, label="whole q-range", alpha=0.5)
    ax2.hist(srr_array[:,3], bins=40, label="small angle only", alpha=0.5)

    ymin, ymax = ax2.get_ylim()
    ax2.set_ylim(ymin, ymax)

    for name, srr in [("20180526/GI", 0.511472), ("20180526/OA", 0.638757), ("20190527/EndoAB", 0.957003)]:
        ax2.plot([srr, srr], [ymin, ymax], ":", label="%s: SRR=%.3g" % (name, srr))

    ax2.legend()
    ax2.set_xlabel("SRR")
    ax2.set_ylabel("Counts")

    if srr_array.shape[1] > 4:
        rg1 = srr_array[:,4]
        rg2 = srr_array[:,5]
        rdr = (rg2 - rg1)*2/(rg1 + rg2)
        ax3.set_title("RDR Distribution", fontsize=16)
        ax3.hist(rdr, bins=40, label="RDR")

        ymin, ymax = ax3.get_ylim()
        ax3.set_ylim(ymin, ymax)

        for name, rdr in [("20180526/GI", -0.034760479), ("20180526/OA", 0.144374836), ("20190527/EndoAB", 0.033599555)]:
            ax3.plot([rdr, rdr], [ymin, ymax], ":", label="%s: RDR=%.3g" % (name, rdr))

        ax3.legend()
        ax3.set_xlabel("RDR")
        ax3.set_ylabel("Counts")

    fig.tight_layout()
    plt.show()

if __name__ == '__main__':
    import os
    import sys
    this_dir = os.path.dirname( os.path.abspath( __file__ ))
    home_dir = os.path.abspath( this_dir + '/..' )
    sys.path.append(home_dir)
    from molass_legacy.KekLib.OurMatplotlib import mpl_1_5_backward_compatible_init
    mpl_1_5_backward_compatible_init()
    # srr_csv = r"D:\TODO\20230904\SRR-save\srr\srr.csv"
    srr_csv = r"D:\TODO\20230904\SRR-save3\srr\srr.csv"
    demo(srr_csv)
